#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)address_type.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991 BY                            *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      address_type.c
**
**  ABSTRACT:
**
**	Grammar element that parses an address type.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**  05-Feb-1999  Shi-Jung Kao
**	Extended to ANSI mode, address type may contain qualification.
**
**  30-Jun-1999  Shi-Jung Kao
**	Fixed the malloc(strlen()), should be malloc(strlen()+1).
**
**  3-Nov-1999	 Shi-Jung Kao
**	Accepted ZnZ, Zm_nZ nontype form.
**
**  09-Feb-2000  Shi-Jung Kao
**	Supported nested template template args.  Z1__tm__4_Z2ZZ
**	Added DEMANGLE_TRACEIN/TRACEOUT.
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

/* language element particular definitions */
#define ASCII_ZERO 48
#if DEC_ANSI_FORM
int /* recognition status */
_address_type(const char* s,	/* IN:  string to be parsed */
	      int start,	/* IN:  index of first unparsed character */
	      int* end,		/* OUT: index of last parsed character */
	      struct MLD_type** t)
				/* OUT: type */
#else
int /* recognition status */
_address_type(const char* s,	/* IN:  string to be parsed */
      		 int start,	/* IN:  index of first unparsed character */
       		 int* end,	/* OUT: index of last parsed character */
       		 char* a,	/* OUT: string to contain address object */
       		 int max_a)	/* IN:	maximum size of a */
#endif
{
  int start_of_address;
  int end_of_address;
  int size_of_address;

  /* variables common to all language elements */
  int ci;  /* current index */
  char ch; /* current character */
  
#if DEC_ANSI_FORM
  struct MLD_unmodified_type* ut = 0L; /* pointer to unmodified type */
  struct MLD_type* t_temp = 0L; /* temporary pointer to type structure */
  struct MLD_address* ad = 0L;
  int end_of_grammar_element;
  int temp_int;
  struct MLD_qualification* q = 0L;
  int status;
  char addr_name[STRSIZ];
#endif

  DEMANGLE_TRACEIN(_address_type);
  ch = s[ci = start]; /* initialize ch and ci */

  /* recognize "_" */
  if (ch != '_')
  {
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_NOTHING_RECOGNIZED;
  } 
  ch = s[++ci]; /* increment ch and ci */
  start_of_address = ci;
  
  /* get letters and numbers */  
  while ( ((ch >= '0') && (ch <= '9')) ||
          ((ch >= 'a') && (ch <= 'z')) || (ch == '$') ||
          ((ch >= 'A') && (ch <= 'Z')) )
  {
      ch = s[++ci]; /* increment ch and ci */
  }
  end_of_address = ci - 1;
  
  /* check for an ending '_' */
  if (ch != '_')
  {
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_NOTHING_RECOGNIZED;
  }
 
  /* The address name must consist of at least one character */
  if (start_of_address > end_of_address)
  {
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_NOTHING_RECOGNIZED;
  }

  /* The address name must NOT begin with a number */
  if ((s[start_of_address] >= '0') && (s[start_of_address] <= '9'))
  {
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_NOTHING_RECOGNIZED;
  }
  
  *t = 0L;

#if DEC_ANSI_FORM
  /* At this point we know we have an address object */

  /*********************************
     Accept both cases:
     f__7C__T_i_Xv
             ^^^
     f__11C__T_i__1S_Xv
              ^^^^^^^
  *********************************/

  if (ci+1 <= strlen(s) && s[ci+1] == '_') {
    /* check if the we're dealing with a qualified nontype */
    status = _parse_qualifications(s, ci+2, &end_of_grammar_element, 
                                (struct MLD_qualification**) &q);
    switch (status) {
    case MLD_SOMETHING_RECOGNIZED:
      /* Check the ending '_' */
      if (end_of_grammar_element < strlen(s) && 
	  s[end_of_grammar_element+1] != '_') {
	_delete_qualifications(q);
	DEMANGLE_TRACEOUT(_address_type);
	return MLD_NOTHING_RECOGNIZED;
      }
      ci = ++end_of_grammar_element;
      break;
    
    case MLD_ERROR:
      ut = 0L;
      *end = start - 1;
      _delete_qualifications(q);
      DEMANGLE_TRACEOUT(_address_type);
      return MLD_ERROR;
      break;

    default:
      break;
    }

  }
  size_of_address = end_of_address - start_of_address + 1;  
  if ((size_of_address + 1) < STRSIZ) {
    strncpy(&addr_name[0], &s[start_of_address], size_of_address);
    addr_name[size_of_address] = 0; /* null terminate */
  } else {
    _delete_qualifications(q);
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_ERROR;  /* NO SPACE !!!! */
  }

  _make_address_type(addr_name, q, t);
  if (!t) {
    DEMANGLE_TRACEOUT(_address_type);
    return MLD_ERROR;
  }
  *end = ci; /* ci is the last digit parsed */
  DEMANGLE_TRACEOUT(_address_type);
  return MLD_SOMETHING_RECOGNIZED;

#else
  /* At this point we know we have a address object, time to copy */
  size_of_address = end_of_address - start_of_address + 1;
  if ((size_of_address + 1) < max_a)
  {
    strncpy(&a[1], &s[start_of_address], size_of_address);
    a[0] = '&'; /* prepend ampersand */
    a[size_of_address+1] = 0; /* null terminate */
  }
  else
  {
    return MLD_ERROR;  /* NO SPACE !!!! */
  }

  *end = ci; /* ci is the last digit parsed */
  DEMANGLE_TRACEOUT(_address_type);
  return MLD_SOMETHING_RECOGNIZED;
#endif
}


void _make_address_type(
	char *addr_name,		/* IN: address name. */
	struct MLD_qualification* q,	/* IN: its qualifier. */
	struct MLD_type** t)		/* OUT: a MLD_type */
{

  struct MLD_unmodified_type* ut = 0L; /* pointer to unmodified type */
  struct MLD_type* t_temp = 0L; /* temporary pointer to type structure */
  struct MLD_address* ad = 0L;

  DEMANGLE_TRACEIN(_make_address_type);
  /* create address type */
  *t = 0L;
  ad = (struct MLD_address*) malloc(sizeof(struct MLD_address));
  if (ad == 0L) {
    _delete_qualifications(q);
    DEMANGLE_TRACEOUT(_make_address_type);
    return;  /* error, no memory allocated */
  }
  DEMANGLE_RECORD_ALLOC(ad);

  ad->qualifications = q;
  ad->type_name = (char*) malloc(strlen(addr_name)+1);
  if (ad->type_name == 0L) {
    _delete_address(ad);
    DEMANGLE_TRACEOUT(_make_address_type);
    return;
  }
  DEMANGLE_RECORD_ALLOC(ad->type_name);

  strcpy(ad->type_name, addr_name);

  /* create unmodified type */
  ut = (struct MLD_unmodified_type*) 
			malloc(sizeof(struct MLD_unmodified_type));
  if (ut == 0L) {
    _delete_address(ad);
    DEMANGLE_TRACEOUT(_make_address_type);
    return;  /* error, no memory allocated */
  }
  (ut)->type_of_type = MLD_ADDRESS_TYPE;
  (ut)->the_type.address = (struct MLD_address*) ad;
  (ut)->shared_id = 0L;
  DEMANGLE_RECORD_ALLOC(ut);

  /* create type structure */
  t_temp = (struct MLD_type*) malloc(sizeof(struct MLD_type)); 
  if (t_temp == 0L) {
    _delete_unmodified_type(ut);
    DEMANGLE_TRACEOUT(_make_address_type);
    return; /* error, no memory allocated */
  }
  DEMANGLE_RECORD_ALLOC(t_temp);
    
  t_temp->ut = ut;
  t_temp->shared_ut = 0;
  t_temp->tm_list = 0L; /* no type modifiers */
  t_temp->next = 0L; /* no next, yet */
  t_temp->repeat_type = 0L;
  t_temp->repeat_count = 0;

  *t = t_temp;
  DEMANGLE_TRACEOUT(_make_address_type);
  return;

}

int /* recognition status */
_nontype_type(const char* s,	/* IN:  string to be parsed */
	      int is_pointer,   /* IN:  is_pointer */
	      int start,	/* IN:  index of first unparsed character */
	      int* end,		/* OUT: index of last parsed character */
	      struct MLD_type** t)
				/* OUT: type */
{
  struct MLD_unmodified_type* ut = 0L; /* pointer to unmodified type */
  struct MLD_type* t_temp = 0L; /* temporary pointer to type structure */
  struct MLD_nontype* nt = 0L;
  struct MLD_type* a_type = 0L;
  struct MLD_mangled_template_name* temp_mn = 0L;
  struct MLD_complex* temp_c;
  char string[STRSIZ];
  int status;
  int end_of_grammar_element;

  DEMANGLE_TRACEIN(_nontype_type);

  string[0] = '\0';
  if (s[start] == 'Z') {
    /* Find ZnZ or Zm_nZ */
    int id;
    status = _number(s, start+1, &end_of_grammar_element, &id);
    if (status == MLD_SOMETHING_RECOGNIZED) {
      end_of_grammar_element++;
      if (s[end_of_grammar_element] == 'Z') {
	;
      } else if (strncmp (&s[end_of_grammar_element], "__tm__", 6) == 0) {

	/* We should call _mangled_template_name in the future. */

	int ci = end_of_grammar_element+6;
	int sn = end_of_grammar_element-start;

	strncpy(&string[0], &s[start], sn);
	
	status = _template_arguments(s, ci, &end_of_grammar_element, &a_type);
	if (s[++end_of_grammar_element] != 'Z') {
	  _delete_type_list(a_type);
	  DEMANGLE_TRACEOUT(_nontype_type);
	  return MLD_NOTHING_RECOGNIZED;
	}
	temp_mn = (struct MLD_mangled_template_name*) 
	  malloc(sizeof(struct MLD_mangled_template_name));
	DEMANGLE_RECORD_ALLOC(temp_mn);
	if (temp_mn == 0L) 
	  return MLD_ERROR; /* error if no memory allocated */
	temp_mn->identifier = 0L;
	temp_mn->args = a_type;
	string[sn] = 'Z';
	string[sn+1] = '\0';

      } else if (s[end_of_grammar_element] == '_') {
	status = _number(s, end_of_grammar_element+1, 
					&end_of_grammar_element, &id);
	if (status == MLD_SOMETHING_RECOGNIZED) {
	  if (s[++end_of_grammar_element] != 'Z') {
	    DEMANGLE_TRACEOUT(_nontype_type);
	    return MLD_NOTHING_RECOGNIZED;  /* Can not be a nontype. */
	  }
	}
      } else {
	DEMANGLE_TRACEOUT(_nontype_type);
	return MLD_NOTHING_RECOGNIZED;  /* Can not be a nontype. */
      }

      if (!temp_mn) {
	/* Normal cases, not a nested nontype. e.g. Z1__tm__4_Z2ZZ */
	strncpy(&string[0], &s[start], end_of_grammar_element-start+1);
	string[end_of_grammar_element-start+1] = '\0';
      }
    }
  } else
    status = _pascal_string(s, start, &end_of_grammar_element, string, STRSIZ);

  if (status == MLD_SOMETHING_RECOGNIZED) {
    /* We should call mangled_variable_name, but that needs structure
     * changes.  We will see how far it can last for now.
     */
    char * du;
    struct MLD_qualification* q=0;/* qualifier */

    /* Case of: 14_f10X__tm__2_i
     *                ^^^^^^^^^^
     * shall be interpreted as: 2 arguments of <float, X<int> >
     * So marked X__tm__2_i is not a nontype here.
     *
     * And if a_type, then this is a template<template...>
     */
    if (strstr(&string[0], "__tm__") && !a_type) {
      DEMANGLE_TRACEOUT(_nontype_type);
      return MLD_NOTHING_RECOGNIZED;  /* Can not be a nontype. */
    }

    if (du = strstr(&string[0], "__")) {
      int length_q = 0;
      status = _parse_qualifications(du, 2, &length_q, &q);
      if (status != MLD_SOMETHING_RECOGNIZED || 
	  length_q+1 != strlen(du)) {
	_delete_qualifications(q);
	q = 0;
      }
      string[du-&string[0]] = '\0';  /* NULL ending. */
    }

    /* create nontype type */
    *t = 0L;
    nt = _alloc_nontype();
    if (nt == 0L) {
      _delete_qualifications(q);
      DEMANGLE_TRACEOUT(_nontype_type);
      return MLD_ERROR;
    }

    nt->is_address = is_pointer; /* a pointer, nontype has to be an address. */
    nt->qualifications = q;
    nt->nontype_name = (char*) malloc(strlen(&string[0])+1);
    if (nt->nontype_name == 0L) {
      _delete_nontype(nt);
      DEMANGLE_TRACEOUT(_nontype_type);
      return MLD_ERROR;
    }
    DEMANGLE_RECORD_ALLOC(nt->nontype_name);

    strcpy(nt->nontype_name, &string[0]);
    nt->nested_args = temp_mn;

    /* create unmodified type */
    ut = (struct MLD_unmodified_type*) 
			malloc(sizeof(struct MLD_unmodified_type));
    if (ut == 0L) {
      _delete_nontype(nt);
      DEMANGLE_TRACEOUT(_nontype_type);
      return MLD_ERROR;
    }
    (ut)->type_of_type = MLD_NONTYPE_TYPE;
    (ut)->the_type.nontype = (struct MLD_nontype*) nt;
    (ut)->shared_id = 0L;
    DEMANGLE_RECORD_ALLOC(ut);

    /* create type structure */
    t_temp = (struct MLD_type*) malloc(sizeof(struct MLD_type)); 
    if (t_temp == 0L) {
      _delete_unmodified_type(ut);
      DEMANGLE_TRACEOUT(_nontype_type);
      return MLD_ERROR;
    }
    DEMANGLE_RECORD_ALLOC(t_temp);
    
    t_temp->ut = ut;
    t_temp->shared_ut = 0;
    t_temp->tm_list = 0L; /* no type modifiers */
    t_temp->next = 0L; /* no next, yet */
    t_temp->repeat_type = 0L;
    t_temp->repeat_count = 0;
    
    *t = t_temp;
    *end = end_of_grammar_element;
    DEMANGLE_TRACEOUT(_nontype_type);
    return MLD_SOMETHING_RECOGNIZED;

  } else {
    *end = start-1;
    *t = 0L;
    DEMANGLE_TRACEOUT(_nontype_type);
    return MLD_NOTHING_RECOGNIZED;
  }

}
