#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)demangle.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991 BY                            *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      demangle.c
**
**  ABSTRACT:
**
**	Demangler accepts input from a file or stdin and passes the
**      input to the demangler function line by line, outputting
**      the demangled output to stdout.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**  06-Feb-1998	 Shi-Jung Kao
**	Add SHOW_MLD_NO_SPACES, SHOW_MLD_DEMANGLED_NAME for developing purpose.
**	Add MLD_DUMP, MLD_MEMCHECK.
**
**  17-Jan-2000  Shi-Jung Kao
**	Add -no_template_details. i.e. no <T0=...>
**
**  04-Feb-2000  Shi-Jung Kao
**	Skip the name that longer than BUFSIZ-2.
**--
**/


/*
// Mangler Filter Program
// ----------------------
// This program does the following:
//  1) accepts a filename or stream
//  2) sends each line to the demangler function
//
// This is all that this program does.  Most of the functionality
// can be found in the demangle and demangler.
*/
#ifdef CXX_NT
#include "dmn_str.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#define SHOW_MANGLED_NAME_IN_PARENS_FLAG	"-show_mangled_name"
#define SHOW_NO_VTABLE_INFO_FLAG		"-no_vtable_info"
#define SHOW_VERSION_FLAG1			"-v"
#define SHOW_VERSION_FLAG2			"-version"
#define SHOW_VERSION_FLAG3			"-ver"

/* The following 3 are for developing purpose. */
#define SHOW_MLD_NO_SPACES			"-mld_no_spaces"
#define SHOW_MLD_DEMANGLED_NAME			"-show_demangled_name"
#define SHOW_MLD_DUMP				"-dump"
#define SHOW_MLD_MEMCHECK			"-memcheck"
#define SHOW_MLD_NO_TEMPLATE_DETAILS		"-no_template_details"

#define ERROR -1
#define NO_FLAGS 0
#define STRSIZE BUFSIZ

int process_arguments(int*nargs,char*argv[]);

main(int nargs, char *argv[])
{
   FILE *in;
   char s[BUFSIZ], ds[STRSIZE];
   char ch;
   char *mangled_file;
   int opts;
   int i,j;
   int status;

   opts = process_arguments(&nargs, argv);                 
   if (opts == ERROR) /* ERROR */
   {
     printf("Usage: %s [-show_mangled_name] [-no_vtable_info] [-] [file...]\n", argv[0]);
     return; /* exit */
   }

   if   ((nargs == 1) /* no file arguments, use stdin */
        /* or "-" is the only argument, use stdin */
     ||  ((nargs == 2) && (argv[nargs-1][0] == '-') && (argv[nargs-1][1] == '\000'))) 
   {
   
#ifdef THINK_C   
     mangled_file = "yyy";
  
     /* open file */
     in = fopen(mangled_file,"r");
     if (!in) 
     {
        printf("%s: No such file or directory\n",mangled_file);
        exit(0);
     }
     
     /* get the opts flag */
     fgets(s,BUFSIZ-1,in); /* get line */
     opts = s[0] - 48;
     
     while(!feof(in))
     {
     /* Step 1:  get a string from the "strings" file */
         fgets(s,BUFSIZ-1,in); /* get line */

     /* Step 2:  demangle the string */
           MLD_demangle_string(s, ds, STRSIZE, opts); /* demangle with options */
       
     /* Step 3:  output the string */
         if (!feof(in)) printf("%s", ds);
     }
#else   
     while(!feof(stdin))
     {
     /* Step 1:  get a string from the "strings" file */
         fgets(s,BUFSIZ-1,stdin); /* get line */

     /* Step 2:  demangle the string */
         status = MLD_demangle_string(s, ds, STRSIZE, opts); /* demangle with options */

/* Uncomment this if you want to know if something is recognize or not 
         switch (status)
         {
           case MLD_SOMETHING_RECOGNIZED:
             printf("Something Recognized\n");
           break;

           case MLD_NOTHING_RECOGNIZED:
             printf("Nothing Recognized\n");
           break;

           case ERROR:
             printf("Error\n");
           break;

           case MLD_INADEQUATE_ARRAY_SIZE:
             printf("Inadequate Array Size\n");
           break;
         }
*/

     /* Step 3:  output the string */
         if (!feof(stdin)) {
	   printf("%s", ds);
	   DEMANGLE_DUMP_ARRAY;
	 }
     }
#endif 
     
   }
   else /* for each file print all of the stuff */
   {
     for (i = 1 ; i < nargs ; i++)
     {
       int skip_until_newline = 0;

       mangled_file = argv[i];
  
       /* open file */
       in = fopen(mangled_file,"r");
       if (!in) 
       {
          printf("%s: No such file or directory\n",mangled_file);
          exit(0);
       }

       /* if there are several files, print the file name */
       if (nargs > 2) /* if there are 2 or more files */
       {
         for(j=0;j<(strlen(argv[i])+4);j++){printf(":");}
         printf("\n: %s :\n",argv[i]);
         for(j=0;j<(strlen(argv[i])+4);j++){printf(":");}
         printf("\n");
       }

       while(!feof(in))
       {
	 char* get_status;

	 /* Step 1:  get a string from the "strings" file */
	 get_status = fgets(s, BUFSIZ-1,in); /* get line */

	 /* Step 2:  demangle the string */
	 if (skip_until_newline || 
	     (strlen(s) == BUFSIZ - 2 && !strstr(s, "\n"))) {
	   skip_until_newline = 1;
	   if (strstr(s, "\n"))
	     skip_until_newline = 0;
	   printf ("%s", s);
	 } else {
	   if (get_status || !feof(in)) {
	     MLD_demangle_string(s, ds, STRSIZE, opts);
	     printf("%s", ds); /* output the string */
	     if (feof(in) && !strstr(s, "\n"))
	       printf("\n"); /* Append newline if necessary. */
	   }
	 }
       }
       DEMANGLE_DUMP_ARRAY;
     }
   }
}

int /* flag value or none (0) or error (-1) */
recognize_flag(char* flag_str, char* next_flag_str)
{
  /* "-" is the file that represents stdin */
  if (!strcmp(flag_str, "-"))
	return NO_FLAGS;

  else if (!strcmp(flag_str, SHOW_MANGLED_NAME_IN_PARENS_FLAG))
        return MLD_SHOW_MANGLED_NAME;

  else if (!strcmp(flag_str, SHOW_NO_VTABLE_INFO_FLAG))
        return MLD_SHOW_INFO; /* actually, don't show info! */

  else if (!strcmp(flag_str, SHOW_MLD_NO_SPACES))
        return MLD_NO_SPACES;

  else if (!strcmp(flag_str, SHOW_MLD_DEMANGLED_NAME)) {
        return MLD_SHOW_DEMANGLED_NAME;
#if !PRODUCTION
  } else if (!strcmp(flag_str, SHOW_MLD_DUMP)) {
        demangle_dump = 1;
        return MLD_DUMP;

  } else if (!strcmp(flag_str, SHOW_MLD_MEMCHECK)) {
        demangle_memcheck = 1;
        return MLD_MEMCHECK;
#endif
  } else if (    (!strcmp(flag_str, SHOW_VERSION_FLAG1))
       || (!strcmp(flag_str, SHOW_VERSION_FLAG2))
       || (!strcmp(flag_str, SHOW_VERSION_FLAG3)))
  {
    printf("This C++ demangler was created:  %s\n", __DATE__);
    exit(1);
  }

  else if (!strcmp(flag_str, SHOW_MLD_NO_TEMPLATE_DETAILS)) {
        demangle_no_template_details = 1;
        return MLD_NO_TEMPLATE_DETAILS;
  }

  else if ((flag_str[0] == '-') || (flag_str[0] == '+'))
	return ERROR;

  /* no flags recognized */
  return NO_FLAGS; 
}

int /* options or -1 for error */
process_arguments(int* nargs, char* argv[])
{
  int opts = MLD_SHOW_DEMANGLED_NAME;
  int opt = NO_FLAGS;
  int number_of_args;
  int i;
  int show_vtable_info = 1; /* on by default */

  number_of_args = *nargs - 1; /* the first argument is the command name */
  *nargs = 1;

#if !PRODUCTION
  demangle_dump = 0;
  demangle_memcheck = 0;
#endif

  demangle_no_template_details = 0;

  for (i = 1 ; i <= number_of_args ; i++)
  {
    if (i<number_of_args) 
	opt = recognize_flag(argv[i], argv[i+1]);
    else
	opt = recognize_flag(argv[i], NULL);

    switch (opt)
    {
      case MLD_SHOW_INFO: /* actually, don't show info */ 
        show_vtable_info = 0; /* turn off */
      break;

      case NO_FLAGS:
        argv[*nargs] = argv[i]; /* copy argument */
        *nargs = *nargs + 1;
      break;
 
      case ERROR:
        return ERROR;
      break;

      default:
        opts = opts | opt; /* add the new flags */
      break;
    }
  }
  if (show_vtable_info) opts = opts | MLD_SHOW_INFO; /* now, finally, 
                                                        add the show
                                                        info flag */
  return opts;
}
