#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)demangle_string.c	1.1      2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991, 1996 BY                      *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      demangle_string.c
**
**  ABSTRACT:
**
**	Demangles a complete string.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
** 05-Jan-1995  Harold Seigel   Check for truncated names and replace last
**                              bunch of characters with 'x's if found one.
**
** 08-Jul-1996  Harold Seigel   Check lengths before calling strcpy to
**                              prevent overwriting the stack.
**
** 12-Aug-1996  Harold Seigel   Move check for truncated names to function
**				MLD_parse_mangled_name.
**
** 27-Oct-1997	Shi-Jung Kao	cxxc_bugs:4782
**	Append '\n' at the output if treat_as_trunc (i.e. truncated string).
**
** 08-Jan-1999	Shi-Jung Kao
**	Added MLD_INIT_INTERNAL_NAME, MLD_FINI_INTERNAL_NAME,
**	MLD_PROMOTED_ENTITY_NAME.
**
** 16-Apr-1999  Shi-Jung Kao	cxxc_bugs:6601
**	Added new decompress_string for __CPRnnnn__.
**
** 18-May-1999  PremAnand M. Rao
**      Added #if 0 around code which was producing 
**      "unreachable" warnings.
**
** 07-Jan-2000  Shi-Jung Kao
**	Accept the leading version7 signature.
**
** 12-Jan-2000  Shi-Jung Kao
**	Accept leading blanks, then follows with __7.
**	Accept the fact of __7__CPR... for compressed names.
**
** 17-Jan-2000  Shi-Jung Kao
**	Accept the compressed mangled name in the middle of the string.
**
** 26-Jan-2000  Shi-Jung Kao
**	Exit the loop when hits the '\n' or beyond the s.
**	Skip the names that longer than BUFSIZ-16.  (too close to the limit)
**	Added DEMANGLE_TRACEIN/TRACEOUT.
**	Take envir variable MLD_DEMANGLE_OPTIONS.
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#include "dmn_str.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

extern int MLD_ias;

/*
 * Indicate what is the version of the demangler, not what form of the
 * demangled names.
 */
unsigned long __MLD_demangle_version = 1;
static char * __MLD_demangle_what = "@(#)Compaq C++ demangling version 0";

int /* status */ 
__MLD_demangle_string(const char* s,	/* IN:  string to be demangled */
       		char* ds,	/* OUT: demangled string */
       		int ds_max,	/* IN:	maximum size of ds */
       		int flags)	/* IN:	flags to control string formation */
{
  int status, end, error_start;
  char bs[BUFSIZ]; /* before string */
  char rs[BUFSIZ]; /* string that is replaced by demangled string */
  char as[BUFSIZ]; /* after string */
  char ts[BUFSIZ]; /* uncompressed string */
  char us[BUFSIZ]; /* temp string */
  char * uu7;
  char * uuCPR;
  char *demangle_options;

  struct MLD_mangled_name* mn; /* mangled name */
  int ret_stat = MLD_NOTHING_RECOGNIZED;
  int number, end_of_number, us_i, ts_i;
  int start=0;

  MLD_ias = 0; /* inadequate array size flag, see utilities.c */

  /* If buffer is not as big as string, it will never work */
  if (strlen(s) >= ds_max) return MLD_INADEQUATE_ARRAY_SIZE;

  /* if flag is equal to just MLD_SHOW_MANGLED_NAME return original string */
  if (flags == MLD_SHOW_MANGLED_NAME) {
    strcpy(ds, s);
    return ret_stat; 
  }

  /* initialize ds to be empty */
    strcpy(ds,"");

  /* Do nothing, if the size is too close to the BUFSIZ.
   * Besides, when a longer name than 8190, the system truncated it,
   * so in this routine the maximum length it can see is 8190.
   */
  if (strlen(s) >= BUFSIZ - 16) {
    strcpy(ds, s);
    return MLD_NOTHING_RECOGNIZED;
  }

  ts_i = 0;
  us_i = 0;
  while (s[us_i] == ' ') {
    us[us_i] = ' ';
    us_i++;
  }
  us[us_i] = '\0';

  strcpy (&ts[0], &s[start]);

  DEMANGLE_TRACEIN(MLD_demangle_string);

  /* Set demangle_no_template_details if necessary. */
  if (demangle_options = getenv("MLD_DEMANGLE_OPTIONS")) {
    if (strstr(demangle_options, "no_template_details"))
      demangle_no_template_details = TRUE;
  }

  /* get first demangled structure */
  status = MLD_parse_mangled_name(&ts[0], 0, &end, &mn, bs, rs, as, BUFSIZ);
  ts_i = end + 1;

  if (status == MLD_SOMETHING_RECOGNIZED) {
    ret_stat = MLD_SOMETHING_RECOGNIZED;
    if (end == strlen(ts) && !mn) {
      /* Found a unnamed case stop here. */

      _concat(ds, ds_max, bs); /* add before string */
      _concat(ds, ds_max, rs);
      _concat(ds, ds_max, as);
      if (flags & MLD_SHOW_MANGLED_NAME) {
        _concat(ds, ds_max, " (");
        _concat(ds, ds_max, &ts[0]);
        _concat(ds, ds_max, ") ");
      }
      if (ts[end-1] == '\n')
        _concat(ds, ds_max, "\n");
      return ret_stat;
    }
  }

  /* If has leading space, and bs has not been filled with space
   * by MLD_parse_mangled_name yet.
   */
  if (us_i > 0 && bs[0] != ' ')
    _concat (ds, ds_max, us);

  while (status == MLD_SOMETHING_RECOGNIZED) {
    while (status == MLD_SOMETHING_RECOGNIZED) {
      _concat(ds, ds_max, bs); /* add before string */
      if (mn->mangled_name_type == MLD_VIRTUAL_TABLE_NAME ||
	  mn->mangled_name_type == MLD_RUNTIMETYPE_INFO_NAME ||
	  mn->mangled_name_type == MLD_INIT_INTERNAL_NAME ||
	  mn->mangled_name_type == MLD_FINI_INTERNAL_NAME) { 
	/* concat rs, if the mangled name is a vt name */
        _concat(ds, ds_max, rs);
      }
      MLD_concat_mangled_name(mn, ds, ds_max, flags);
      if (mn->mangled_name_type == MLD_PROMOTED_ENTITY_NAME)
	_concat(ds, ds_max, as);

      if ((flags & MLD_SHOW_MANGLED_NAME) &&
          (mn->mangled_name_type != MLD_VIRTUAL_TABLE_NAME)) {
        _concat(ds, ds_max, " (");
        _concat(ds, ds_max, rs);
        _concat(ds, ds_max, ") ");
      }

      if (treat_as_trunc)
	_concat(ds, ds_max, "\n");

      /* get next demangled structure */
      MLD_delete_mangled_name(mn); /* done with old mangled name */

      if (end+2 < strlen(ts)) {
	while (ts[ts_i] == ' ') {
	  _concat (ds, ds_max, " ");
	  ts_i++;
	}
	status = MLD_parse_mangled_name(&ts[ts_i], 0, &end, &mn, 
					bs, rs, as, BUFSIZ);
	if (status == MLD_NOTHING_RECOGNIZED)
	  _concat(ds, ds_max, bs); /* end of string reached */
      } else {
	if (ts[end+1] == '\n' && !treat_as_trunc)
	  _concat(ds, ds_max, "\n");
	status = MLD_NOTHING_RECOGNIZED;
      }

      if (status == MLD_SOMETHING_RECOGNIZED) 
	ret_stat = MLD_SOMETHING_RECOGNIZED;
      ts_i += end + 1;
    }

    /* was nothing recognized, then exit */
    if (status == MLD_NOTHING_RECOGNIZED) {
      if (MLD_ias) {
	DEMANGLE_TRACEOUT(MLD_demangle_string);
	return MLD_INADEQUATE_ARRAY_SIZE; 
      } else {
	DEMANGLE_TRACEOUT(MLD_demangle_string);
	return ret_stat;
      }
    }

error:  
    /* did an error occur */
    if (status == MLD_ERROR) { /* if nothing recognized */
      DEMANGLE_TRACEOUT(MLD_demangle_string);
      return MLD_ERROR; /* just return an error--don't bothing 
                         trying to recover.  (This is new code
                         that I've added--I might change it later) */

    }
  }  

  if (strlen(ts) < ds_max)
    strcpy(ds,ts); /* copy if unsuccessful */
  if (MLD_ias) {
    DEMANGLE_TRACEOUT(MLD_demangle_string);
    return MLD_INADEQUATE_ARRAY_SIZE;
  } else {
    DEMANGLE_TRACEOUT(MLD_demangle_string);
    return ret_stat;
  }
}
