#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)type.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991 BY                            *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      type.c
**
**  ABSTRACT:
**
**	Grammar element that parses a type.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**  29-Oct-1997	Shi-Jung Kao
**	Add ptr_to_function_type to MLD_function to handle (*(*)())
**
**  11-Jan-2000 Shi-Jung Kao
**	In _repeat_arg_type, has to consider the repeat_count when
**	walk through the arg lists.
**
**  25-Jan-2000 Shi-Jung Kao
**	First check if start beyond the s, then return nothing_recog.
**	Added DEMANGLE_TRACEIN/TRACEOUT.
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

/* language element particular definitions */
#define ASCII_ZERO 48

int /* recognition status */
_type(const char* s,		/* IN:  string to be parsed */
     int start,			/* IN:  index of first unparsed character */
     int* end,			/* OUT: index of last parsed character */
     struct MLD_type** t)	/* OUT: type */
{
  /* variables particular to this language element */
  int status;
  int end_of_grammar_element;
  char temp_str[STRSIZ];
  struct MLD_type_modifiers* tml = NULL;
  struct MLD_type_modifiers* temp_tml;
  struct MLD_unmodified_type* ut; /* pointer to unmodified type */
  struct MLD_type* t_temp; /* temporary pointer to type structure */

/*
***
*** Get type qualifiers (if any)
***
*/

  *end = start - 1;
  *t = 0L;
  if (start > strlen(s)) {
    return MLD_NOTHING_RECOGNIZED;
  }

  DEMANGLE_TRACEIN(_type);  

  status = _type_modifiers(s, start, &end_of_grammar_element, &tml);
  if (status == MLD_ERROR)
  {
    *end = start - 1;
    *t = 0L;
    DEMANGLE_TRACEOUT(_type);  
    return MLD_ERROR;
  }  
  
/*
***
*** Get unmodified type
***
*/

  status = _unmodified_type(s, end_of_grammar_element + 1, 
			    &end_of_grammar_element, &ut);
  switch (status)
  {
    case MLD_SOMETHING_RECOGNIZED:
      /* allocate storage for the type structure */
      t_temp = (struct MLD_type*) malloc(sizeof(struct MLD_type)); 
      DEMANGLE_RECORD_ALLOC(t_temp);
      if (t_temp == 0L) {
	DEMANGLE_TRACEOUT(_type);
	return MLD_ERROR; /* error, no memory allocated */
      }

      t_temp->next = 0L; /* there is no "next" */
      t_temp->ut = ut;
      t_temp->shared_ut = 0;
      t_temp->tm_list = NULL;
      t_temp->repeat_type = 0L;
      t_temp->repeat_count = 0;

      /*
       * For pointers and references, switch the type modifiers for the
       * type with the type modifiers for the pointer.
       */
      if (ut->type_of_type == MLD_POINTER_TYPE) {
	t_temp->tm_list = ut->the_type.pointer->type_modified->tm_list;
	ut->the_type.pointer->type_modified->tm_list = tml;
      } else if (ut->type_of_type == MLD_FUNCTION_TYPE) {
	ut->the_type.function->tm_list = tml;
      } else {
	if (ut->type_of_type == MLD_REFERENCE_TYPE) {
	  t_temp->tm_list = ut->the_type.reference->type_modified->tm_list;
	  ut->the_type.reference->type_modified->tm_list = tml;
	  if (tml == NULL && t_temp->tm_list == NULL &&
	      ut->the_type.reference->type_modified->ut->type_of_type
							== MLD_POINTER_TYPE) {
	    ut->the_type.reference->type_modified->tm_list = 
	      ut->the_type.reference->type_modified->ut->the_type.pointer->
						type_modified->tm_list;
	    ut->the_type.reference->type_modified->ut->the_type.pointer->
						type_modified->tm_list = NULL;
	    /* Promote the tml to the reference. */
	    if (t_temp->tm_list == NULL) {
	      t_temp->tm_list = ut->the_type.reference->type_modified->tm_list;
	      ut->the_type.reference->type_modified->tm_list = NULL;
	    }
	  }
	} else {
	  if (ut->type_of_type == MLD_POINTER_TO_MEMBER_TYPE) {
	    t_temp->tm_list =
	      ut->the_type.ptr_to_member->type_modified->tm_list;
	    ut->the_type.ptr_to_member->type_modified->tm_list = tml;
	  } else {
	    t_temp->tm_list = tml;
	  }
	}
      }
      
      *t = t_temp;
      *end = end_of_grammar_element;
      DEMANGLE_TRACEOUT(_type);  
      return MLD_SOMETHING_RECOGNIZED;
    break;
    
    case MLD_ERROR:
      /* if there's a type modifier list, deallocate storage */
      _delete_type_modifier_list(tml);
    
      *t = 0L;
      *end = start - 1;
      DEMANGLE_TRACEOUT(_type);  
      return MLD_ERROR;
    break;

    default:
    case MLD_NOTHING_RECOGNIZED:
      if (tml != 0L) /* if type modifiers have been recognized... */
      {              /* ...then there should have been a type! Nothing found!*/
        /* delete existing type modifiers */
        _delete_type_modifier_list(tml);
      }
      *t = 0L;
      *end = start - 1;
      DEMANGLE_TRACEOUT(_type);  
      return MLD_NOTHING_RECOGNIZED;
    break;
  }
}

int /* recognition status */
_repeat_arg_type(
      struct MLD_type* arg1,	/* IN:  Head of the type list */
      const char* s,		/* IN:  string to be parsed */
      int start,		/* IN:  index of first unparsed character */
      int* end,			/* OUT: index of last parsed character */
      struct MLD_type** t)	/* OUT: type */
{
  int status;
  int end_of_grammar_element;
  char temp_str[STRSIZ];
  struct MLD_type* t_temp; /* temporary pointer to type structure */
  struct MLD_type* t_temp2; /* temporary pointer to type structure */
  int i,count;

  DEMANGLE_TRACEIN(_repeat_arg_type);

  status = _number(s, start, &end_of_grammar_element, &count);
  if (status == MLD_ERROR || status == MLD_NOTHING_RECOGNIZED) {
    *t = 0L;
    *end = start - 1;
    DEMANGLE_TRACEOUT(_repeat_arg_type);  
    return status;
  }
  i = 1;
  for (t_temp = arg1; t_temp && i<count; i++,t_temp=t_temp->next) {
    if (t_temp->repeat_count)
      i += t_temp->repeat_count - 1;
    continue;
  }
  if (t_temp && i == count) {
    /* allocate storage for the type structure */
    t_temp2 = (struct MLD_type*) malloc(sizeof(struct MLD_type)); 
    DEMANGLE_RECORD_ALLOC(t_temp2);
    if (t_temp2 == 0L) {
      *t = 0L;
      *end = start - 1;
      DEMANGLE_TRACEOUT(_repeat_arg_type);  
      return MLD_ERROR; /* error, no memory allocated */
    }
 
    t_temp2->next = 0L; /* there is no "next" */
    t_temp2->ut = 0L;
    t_temp2->shared_ut = 0;
    t_temp2->tm_list = NULL;
    t_temp2->repeat_type = t_temp;
    t_temp2->repeat_count = 0;
    
    *t = t_temp2;  
    *end = end_of_grammar_element;
    DEMANGLE_TRACEOUT(_repeat_arg_type);  
    return MLD_SOMETHING_RECOGNIZED;
  }

  *t = 0L;
  *end = start - 1;
  DEMANGLE_TRACEOUT(_repeat_arg_type);  
  return MLD_NOTHING_RECOGNIZED;
}
