#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)mangled_name.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991, 1996 BY                      *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      mangled_name.c
**
**  ABSTRACT:
**
**	Grammar element that parses a mangled_name.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**	05-Jan-1995  Harold Seigel
**	Check for truncated names and replace last bunch of characters with
**	'x's if found one.
**
**	18-Nov-1997  Shi-Jung Kao
**	Accept the case that mangled name has length 1022 and not CRCed.
**
**	08-Jan-1999  Shi-Jung Kao
**	Process promoted entity.  i.e. __Ln__ and __LSG__ cases.
**
**	14-Apr-1999  Shi-Jung Kao	cxxc_bugs:6625
**	Early return from processing promoted entity causes regression on
**	internal local class function names.
**
**	27-Dec-1999  Shi-Jung Kao
**	Check if MLD_VERSION7...
**
**      07-Jan-2000  Shi-Jung Kao
**      Move of code checking version7 to demangle_string.c
**
**	12-Jan-2000  Shi-Jung Kao
**	Adjust the treat_as_trunc accordingly due to "__7" extra prefix.
**
**	24-Jan-2000  Shi-Jung Kao
**	Adjust the value that related to TRUNC_LENGTH accordingly if is
**      version7 because it is defined as: MAX_NAME_LENGTH - CRC_LENGTH.
**	Initial MLD_mangled_version here, in case its called at the top.
**
**	26-Jan-2000  Shi-Jung Kao
**	Accept a truncated name length greater than MAX_NAME_LENGTH.
**	Reset shared_id in every call of MLF_parse_mangled_name.
**	This may cause the demangled template have different nontype
**	names from cplus-dem.c for ecc in linux.
**	Skip the names that length too close to the BUFSIZ.
**	Restored CRC back to str_replaced_by_name before return.
**	Perform "decompressing" on the fly.
**	Added DEMANGLE_TRACEIN/TRACEOUT/INIT/FINI.
**	Added process_str_replaced_by_name_before_return.
**	Accept stand alone unnamed...
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#define MAX_GEM_NAME_LENGTH 1024

/*
 * If true, then treat the string as a truncated string.  This means that the
 * last 7 characters have been hashed.
 *
 * Note that C++ V5.x template instantiation depends on treat_as_trunc being
 * set by MLD_parse_mangled_name.  It also depends on the fact that
 * MLD_parse_mangled_name does not clear treat_as_trunc on exit.  Do not
 * change this behavior !!!  (See C++ V5.x driver module deccxx_inst.c for
 * more info.)
 */
long treat_as_trunc;
enum MLD_version MLD_mangled_version = MLD_PRE_VERSION7;
static int /* TRUE or FALSE */
decompress_string(const char* s,	/* IN: string to be decompressed. */
		  int         start,	/* IN: start position of s. */
		  int*        end,	/* IN: start position of s. */
		  char*       ds,	/* OUT: decompressed string. */
		  int ds_max);		/* IN: maximum size of ds. */
static void
process_str_replaced_by_name_before_return(
	   char *str_replaced_by_name,	/* IN/OUT: To be checked. */
	   int   max_str,		/* max size of str_replaced... */
	   char *crcstr,		/* Original CRCed 7 characters. */
	   int   treat_as_truncation,	/* See if we need put crc back. */
	   int   version7_signature);	/* See if prepend "__7" is necessary */
		   

int /* recognition status */
MLD_parse_mangled_name
  (const char* orig_s,		/* IN:  string to be parsed */
   int start,			/* IN:  index of first unparsed char */
   int* end,			/* OUT: index of last parsed char */
   struct MLD_mangled_name** mn,/* OUT: structure containing mangled name */
   char* str_before_name,	/* OUT: string before mangled name */
   char* str_replaced_by_name,	/* OUT: string replaced by mangled name */
   char* str_after_name,	/* OUT: string after mangled name */
   int size_of_string)		/* IN:	max size of the strings listed above */
{
  /* variables common to all language elements */
  int ci;  /* current index */
  char ch; /* current character */
  
  /* variables particular to this language element */
  int status;
  int end_of_grammar_component;
  struct MLD_mangled_name* temp_mn;
  struct MLD_mangled_function_name* mfn;
  struct MLD_mangled_template_name* mtn;
  struct MLD_mangled_variable_name* mvn; 
  enum MLD_mangled_name_types mld_type;
  char info[STRSIZ]; 
  char s[BUFSIZ]; /* before string with possible 'x's */
  char crcstr[10];  /* CRC has 7 characters, use 10 for GEM truncation. */

  /* variables for handling __CPR compressed names */
  char ts[BUFSIZ]; /* uncompressed string */
  char us[BUFSIZ]; /* temp string */
  int decompressed = FALSE;
  int number, end_of_number;
  char * uuCPR;
  char * uu7;
  int adj_s_end;
  int size_upto_uuCPR = 0;
  int version7_signature = 0;

  long orig_s_len, adj_s_len = 0;

  /* adj_s_len: is due to the call to MLD_demangle_string first.
   * The "__7" has been stripped off the original mangled string.
   * We have to check "orig_s_len+adj_s_len" in order to verify 
   * if this is a CRCed/truncated name.
   * i.e. mangled name when call MLD_demangle_string is:
   *	"__7foo__tm__2_i__Fv"
   * Now, we have: adj_s_len = 3, and orig_s
   *	"foo__tm__2_i__Fv"
   */

  char* uuLn;
  int ansi_template;
  int version_start = 0;
  int treat_as_gem_truncated = 0;

  /* initialization particular to language element */
  temp_mn = 0L;
  *mn = 0L;
  *end = start - 1;
  orig_s_len = strlen(orig_s);

  /* Do nothing if the size is too close to BUFSIZ or
   * the start is beyond the length.
   */
  if (start >= orig_s_len || orig_s_len >= BUFSIZ - 16) {
    return MLD_NOTHING_RECOGNIZED;
  }

  _init_shared_id();

#if !PRODUCTION
  {
    static int init_once = 0;
  
    if (!init_once) {
      demangle_memcheck = 1;
      // demangle_trace_enable = 1;
      DEMANGLE_INIT;
      init_once = 1;
    }
  }
#endif

  DEMANGLE_TRACE_INIT;
  DEMANGLE_TRACEIN(MLD_parse_mangled_name);

  /*
   * See if the string should be treated as a truncated string by looking
   * at its length.
   */

  while (orig_s[version_start] == ' ')
    version_start++;

  /*
   * start: index points to the first unparsed char.
   */
  if (strncmp(&orig_s[version_start], VERSION7, VERSION_STRING_LENGTH) == 0) {
    /* If someone calls this routine directly. */
    version7_signature = TRUE;
    MLD_mangled_version = MLD_VERSION7;
    if (start < version_start+VERSION_STRING_LENGTH)
      start = version_start+VERSION_STRING_LENGTH;
  } else
    adj_s_len = (MLD_mangled_version == MLD_PRE_VERSION7) ? 
					0 : VERSION_STRING_LENGTH;

  /* Copy orig_s to us, replace all " __7" with " ". */
  if (uu7 = strstr (&orig_s[0], " __7")) {
    version7_signature = TRUE;
    MLD_mangled_version = MLD_VERSION7;
    /* First copy from 0 to the blank before VERSION7. */
    strncpy (&us[0], &orig_s[0], uu7-&orig_s[0]+1);
    strcpy (&us[uu7-&orig_s[0]+1], uu7+VERSION_STRING_LENGTH+1);
    while (uu7 = strstr(uu7+VERSION_STRING_LENGTH+1, " __7")) {
      char *uu7us = strstr(&us[0], " __7");
      strcpy (uu7us+1, uu7+VERSION_STRING_LENGTH+1);
    }
  } else {
    /* "us" is used only for compressed mangled names, if it ever used. */
    strcpy (&us[0], &orig_s[start]);
  }

  /* Check if is compressed mangled name string.
   * __CPRnnnn__, where nnnn represents length of original mangled name.
   */
  if (strncmp(&us[0], "__CPR", 5) == 0) {
    /* An obvious leading __CPR, compressed mangled name. */
    status = _number(&us[0], 5, &end_of_number, &number);
    if (status == MLD_SOMETHING_RECOGNIZED && 
	us[end_of_number+1] == '_' && us[end_of_number+2] == '_') {

      int  ci = end_of_number+3;  /* Advance beyond signature of compressing */

      /* Decompress string to "ts" */
      if (number < BUFSIZ/2 && 
	  decompress_string(&us[0], ci, &adj_s_end, &ts[0], BUFSIZ)) {
	int ts_size = strlen(ts);
	/* Allow a carrage return at the end of ts string. */
	decompressed = (number == ts_size || 
			(number+1 == ts_size && ts[number] == '\n'));
	if (decompressed) {
	  start = 0;  /* The decompressed string has to start from 0 here. */
	  size_upto_uuCPR = version7_signature? VERSION_STRING_LENGTH : 0;
	}
      }
      if (number >= BUFSIZ/2) {
	DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
	DEMANGLE_TRACE_FINI;
	return MLD_NOTHING_RECOGNIZED;
      }
    }
  } else if (uuCPR = strstr (&us[0], " __CPR")) {
    /* Somewhere in the middle of the string, comes a compressed name. */

    strncpy (&ts[0], &us[0], uuCPR-&us[0]+1);
    status = _number(uuCPR+1, 5, &end_of_number, &number);
    /* Note, uuCPR points at the blank before "__CPR". */

    if (status == MLD_SOMETHING_RECOGNIZED && 
	uuCPR[end_of_number+2] == '_' && uuCPR[end_of_number+3] == '_') {
      int  ci = end_of_number+3;  /* Advance beyond signature of compressing */

      size_upto_uuCPR = uuCPR-&us[0]+1;

      /* Decompress string to "us" */
      if (number < BUFSIZ/2 && 
	  decompress_string(uuCPR+1, ci, &adj_s_end, 
			    &ts[size_upto_uuCPR], BUFSIZ)) {
	int ts_size = strlen(ts);

	/* Allow a carrage return at the end of ts string. */
	decompressed = (number+size_upto_uuCPR == ts_size || 
			(number+size_upto_uuCPR+1 == ts_size && 
			 ts[number+size_upto_uuCPR] == '\n'));
      }
      if (number >= BUFSIZ/2) {
	DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
	DEMANGLE_TRACE_FINI;
	return MLD_NOTHING_RECOGNIZED;
      }
    }
  }

  /*
   * Make a copy of the original string and replace the last 7 characters
   * with 'x's, if the string is one that will be treated as a truncated
   * string.
   *
   * There are 2 kinds valid truncated mangled name:
   * 1. length of 1022, the last 7 characters are CRCed.
   * 2. length of 1024, GEM truncated names.
   */
  if (decompressed)
        strcpy (s, ts);
  else if (orig_s_len < BUFSIZ)
        strcpy (s, orig_s);

  treat_as_gem_truncated = !strstr (orig_s, " ") && 
			(((orig_s_len + adj_s_len == MAX_GEM_NAME_LENGTH &&
			   orig_s[orig_s_len-1] != '\n') ||
			  (orig_s_len + adj_s_len == MAX_GEM_NAME_LENGTH+1 &&
			   orig_s[orig_s_len-1] == '\n')));

  treat_as_trunc = treat_as_gem_truncated ||
		((orig_s_len + adj_s_len == MAX_NAME_LENGTH) || 
		 ((orig_s_len + adj_s_len ==	MAX_NAME_LENGTH + 1) &&
		  (orig_s[orig_s_len - 1] == '\n')));


  if (treat_as_trunc) {
    int i = (treat_as_gem_truncated ? MAX_GEM_NAME_LENGTH : TRUNC_LENGTH)
								- adj_s_len;
    int j,power10,pascal_string_len;
    char ch[2];

    crcstr[0] = '\0';
    /*
     * Search backward, if there is a pascal string like. e.g. "3foo"
     * We want to check if the mangled name has no CRC at the end.
     */
    ch[1] = '\0';
    while (s[i] < '0' || s[i] > '9')
      i--;
    j = (treat_as_gem_truncated ? MAX_GEM_NAME_LENGTH : TRUNC_LENGTH)
							- adj_s_len - i;
    ch[0] = s[i];
    pascal_string_len = atoi(ch);
    power10 = 10;
    while (s[i-1] >= '0' && s[i-1] <= '9') {
      ch[0] = s[i-1];
      pascal_string_len = pascal_string_len + atoi(ch)*power10;
      power10 *= 10;
      i--;
    }
    /*
     * Guess if this is a mangled name without a CRC at the end.
     * Otherwise, replace the CRC with xxxxxxx.
     * To make it simple, do the same to treat_as_gem_truncated.
     */
    if (pascal_string_len != j+6) {
      int tail_len = treat_as_gem_truncated? 9 : 7;
      strncpy (&crcstr[0], s+TRUNC_LENGTH-adj_s_len, tail_len);
      crcstr[tail_len] = '\0';
      strcpy (s+TRUNC_LENGTH-adj_s_len, "xxxxxxx");
    } else
      treat_as_trunc = FALSE;
  }

  ch = s[ci = start]; /* initialize ci and ch */

  /*
  ** Parse a stand alone unnamed .., exclude the __V of
  ** unnamed member variable.
  */

  if (strncmp(s, UNNAMED_MEMBER_VARIABLE_TOKEN, UNNAMED_TOKEN_LEN) != 0
      && demangle_unnamed(&s[start], str_replaced_by_name) ==
                                         MLD_SOMETHING_RECOGNIZED) {
    if (decompressed)
      *end = start + adj_s_end;
    else
      *end = strlen(s);
    strcpy(str_before_name, "");
    strcpy(str_after_name, "");
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_SOMETHING_RECOGNIZED;
  }
  /*
  **
  ** Try to recognize a promoted entity.
  **
  */
  if (uuLn = strstr(&s[start], "__L")) {
    //
    // Check if is promoted entity case: name__Ln__MangledFunctionName.

    char bs[STRSIZ];		/* string before demangled string */
    char id[STRSIZ];		/* identifier string */
    struct MLD_mangled_function_name* mfn;
  
    if (strncmp(&s[start], "__LSG__", 7) == 0)
      ci = ci + 7;
    status = _identifier(s, ci, &end_of_grammar_component, 1,
						id, STRSIZ, bs, STRSIZ);
    if (status != MLD_SOMETHING_RECOGNIZED) {
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return status;
    }
    ch = s[ci = (end_of_grammar_component+1)]; /* update ci and ch */
    if ((s[ci] == '_') && (s[ci+1] == '_') && (s[ci+2] == 'L') &&
	  s[ci+3] >= '0' && s[ci+3] <= '9') {

      ch = s[ci = (ci+4)]; /* update ci and ch */
      while (ch >= '0' && ch <= '9')
	ch = s[++ci];
      if (s[ci] == '_' && s[ci+1] == '_') {

	ci = ci+2;
	status = _mangled_function_name(s, ci, &end_of_grammar_component, &mfn, 
	str_before_name, str_replaced_by_name, str_after_name, size_of_string);

	strcpy(str_after_name, "::"); /* Overwrite the "\n". */
	_concat(str_after_name, size_of_string, id);

	if (decompressed)
	  *end = start + adj_s_end;
	else
	  *end = end_of_grammar_component;
	
	temp_mn = (struct MLD_mangled_name*)
			malloc(sizeof(struct MLD_mangled_name));
	DEMANGLE_RECORD_ALLOC(temp_mn);
	if (temp_mn == 0L) {
	  _delete_mangled_function_name(mfn);
	  DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
	  DEMANGLE_TRACE_FINI;
	  return MLD_ERROR; /* error if no memory allocated */
	}
      
	temp_mn->mangled_name_type = MLD_PROMOTED_ENTITY_NAME;
	temp_mn->name.funct_name = mfn;
      
	*mn = temp_mn;

	process_str_replaced_by_name_before_return (str_replaced_by_name,
	    size_of_string, &crcstr[0], treat_as_trunc, version7_signature);

	DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
	DEMANGLE_TRACE_FINI;
	return MLD_SOMETHING_RECOGNIZED;
      }
    }
  }

  ch = s[ci = start]; /* reset ci and ch */

  /*
  **
  ** Try to recognize a virtual table name
  **
  */

  status = _virtual_table_name(
       s, ci, &end_of_grammar_component, info, str_before_name,
       str_replaced_by_name, str_after_name, size_of_string, &mld_type);
  switch (status) {    
  case MLD_SOMETHING_RECOGNIZED:
    if (decompressed) {
      *end = start + size_upto_uuCPR + adj_s_end - 1;
      if (orig_s[*end] == '\n')
	(*end)--;  /* Backup from '\n'. */
    } else {
      if (end_of_grammar_component == (MAX_NAME_LENGTH-1) &&
	  treat_as_gem_truncated)
	*end = MAX_GEM_NAME_LENGTH -1;
      else
	*end = end_of_grammar_component;
    }
      
    temp_mn = (struct MLD_mangled_name*)malloc(sizeof(struct MLD_mangled_name));
    DEMANGLE_RECORD_ALLOC(temp_mn);
    if (temp_mn == 0L) {
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return MLD_ERROR; /* error if no memory allocated */
    }
      
    temp_mn->mangled_name_type = mld_type;
      
    temp_mn->name.info = (char*) malloc(strlen(info)+1);
    DEMANGLE_RECORD_ALLOC(temp_mn->name.info);
    if (temp_mn->name.info == 0L) {
      DEMANGLE_FREE(temp_mn);
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return MLD_ERROR;
    }
    strcpy(temp_mn->name.info, info);
      
    *mn = temp_mn;

    process_str_replaced_by_name_before_return (str_replaced_by_name,
	size_of_string, &crcstr[0], treat_as_trunc, version7_signature);

    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_SOMETHING_RECOGNIZED;
    break;
    
  case MLD_ERROR:            
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_ERROR;
    break;

  default:
    break;
  }
 
  ansi_template = (strstr(s, "__tm__") != NULL);
  status = MLD_NOTHING_RECOGNIZED;

  /*
  **
  ** Try to recognize mangled function name
  **
  */

  status = _mangled_function_name(
	  s, ci, &end_of_grammar_component, &mfn, str_before_name,
	  str_replaced_by_name, str_after_name, size_of_string);
  switch (status) {    
  case MLD_SOMETHING_RECOGNIZED:
    if (decompressed) {
      *end = start + size_upto_uuCPR + adj_s_end - 1;
      if (orig_s[*end] == '\n')
	(*end)--;  /* Backup from '\n'. */
    } else {
      if (end_of_grammar_component == (MAX_NAME_LENGTH-1) &&
	  treat_as_gem_truncated)
	*end = MAX_GEM_NAME_LENGTH -1;
      else
	*end = end_of_grammar_component;
    }
      
    temp_mn = (struct MLD_mangled_name*)malloc(sizeof(struct MLD_mangled_name));
    DEMANGLE_RECORD_ALLOC(temp_mn);
    if (temp_mn == 0L) {
      _delete_mangled_function_name(mfn);
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return MLD_ERROR; /* error if no memory allocated */
    }
      
    temp_mn->mangled_name_type = MLD_FUNCTION_MANGLED_NAME;
    temp_mn->name.funct_name = mfn;
      
    *mn = temp_mn;

    process_str_replaced_by_name_before_return (str_replaced_by_name,
	size_of_string, &crcstr[0], treat_as_trunc, version7_signature);

    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_SOMETHING_RECOGNIZED;
    break;
    
  case MLD_ERROR:            
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_ERROR;
    break;

  default:
    break;
  }
  

  /*
  **
  ** Try to recognize mangled variable name
  **
  */

  status = _mangled_variable_name(
	  s, ci, &end_of_grammar_component, &mvn, str_before_name,
	  str_replaced_by_name, str_after_name, size_of_string);
  switch (status) {    
  case MLD_SOMETHING_RECOGNIZED:
    if (decompressed) {
      *end = start + size_upto_uuCPR + adj_s_end - 1;
      if (orig_s[*end] == '\n')
	(*end)--;  /* Backup from '\n'. */
    } else {
      if (end_of_grammar_component == (MAX_NAME_LENGTH-1) &&
	  treat_as_gem_truncated)
	*end = MAX_GEM_NAME_LENGTH -1;
      else
	*end = end_of_grammar_component;
    }
      
    temp_mn = (struct MLD_mangled_name*) malloc(sizeof(struct MLD_mangled_name));
    DEMANGLE_RECORD_ALLOC(temp_mn);
    if (temp_mn == 0L) {
      _delete_mangled_variable_name(mvn);
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return MLD_ERROR; /* error if no memory allocated */
    }
      
    temp_mn->mangled_name_type = MLD_VARIABLE_MANGLED_NAME;
    temp_mn->name.var_name = mvn;
      
    *mn = temp_mn;

    process_str_replaced_by_name_before_return (str_replaced_by_name,
	size_of_string, &crcstr[0], treat_as_trunc, version7_signature);

    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_SOMETHING_RECOGNIZED;
    break;
    
  case MLD_ERROR:            
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_ERROR;
    break;

  default:
    break;
  }
  
  
  /*
  **
  ** Try to recognize mangled template name
  **
  */

  status = _mangled_template_name(
	  s, ci, &end_of_grammar_component, &mtn, str_before_name, 
	str_replaced_by_name, str_after_name, size_of_string);
  switch (status) {    
  case MLD_SOMETHING_RECOGNIZED:
    if (decompressed) {
      *end = start + size_upto_uuCPR + adj_s_end - 1;
      if (orig_s[*end] == '\n')
	(*end)--;  /* Backup from '\n'. */
    } else {
      if (end_of_grammar_component == (MAX_NAME_LENGTH-1) &&
	  treat_as_gem_truncated)
	*end = MAX_GEM_NAME_LENGTH -1;
      else
	*end = end_of_grammar_component;
    }
      
    temp_mn = (struct MLD_mangled_name*) malloc(sizeof(struct MLD_mangled_name));
    DEMANGLE_RECORD_ALLOC(temp_mn);
    if (temp_mn == 0L) {
      _delete_mangled_template_name(mtn);
      DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
      DEMANGLE_TRACE_FINI;
      return MLD_ERROR; /* error if no memory allocated */
    }
      
    temp_mn->mangled_name_type = MLD_TEMPLATE_MANGLED_NAME;
    temp_mn->name.templ_name = mtn;
      
    *mn = temp_mn;
//    if (treat_as_trunc && crcstr[0] != '\0' && 
//	strncmp(str_replaced_by_name+strlen(str_replaced_by_name)-7, 
//		"xxxxxxx", 7) == 0)
//      strcpy (str_replaced_by_name+strlen(str_replaced_by_name)-7, crcstr);
    process_str_replaced_by_name_before_return (str_replaced_by_name,
	size_of_string, &crcstr[0], treat_as_trunc, version7_signature);
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_SOMETHING_RECOGNIZED;
    break;
    
  case MLD_ERROR:            
    DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
    DEMANGLE_TRACE_FINI;
    return MLD_ERROR;
    break;

  default:
    break;
  }
 

  /*
  **
  ** At this point, nothing has been recognized
  **
  */
  DEMANGLE_TRACEOUT(MLD_parse_mangled_name);
  DEMANGLE_TRACE_FINI;
  return MLD_NOTHING_RECOGNIZED;
}

static int  /* TRUE or FALSE */
decompress_string(const char* s,	/* IN: string to be decompressed. */
		  int         start,	/* IN: start position of s. */
		  int*	      end,	/* OUT: end position of s. */
		  char*       ds,	/* OUT: decompressed string. */
		  int max_ds)		/* IN: maximum size of ds. */
{
  int si,dsi = 0;  /* index of string s and ds. */
  int slen = strlen(s);
  char ch = s[0];

  /* Decode all "JnnnJ",  where "nnn" is the position of
   * the previous identical string.
   * And case of "JJ" for original single "J".
   * Note, get to the null as well i.e. si == slen.
   */
  *end = start - 1;
  for (si = start; si <= slen; si++) {
    if (s[si] == 'J') {
      if (s[si+1] == 'J') { /* an original single 'J' */
	ds[dsi++] = 'J';
	si++;  /* Advance the extra 'J' in s. */
      } else {  /* Looking for JnnnJ */
	int end_of_number,end_of_number2; /* Position index in s */
	int nnn,len2;
	int status;
	status = _number(s, si+1, &end_of_number, &nnn);
	if (status != MLD_SOMETHING_RECOGNIZED || s[end_of_number+1] != 'J')
	  return FALSE;
	/* Interpret JnnnJ, copy string at ds[nnn] of length len2+ over.
	 * The string at ds[nnn] must have form of "mmString" where mm
	 * represents the size of String.
	 * The actual size of string to be copied is the size of "mmString"
	 * at ds[nnn]. i.e. len2+end_of_number2-nnn+1
	 */
	status = _number(ds, nnn, &end_of_number2, &len2);
	if (status != MLD_SOMETHING_RECOGNIZED)
	  return FALSE;
	strncpy(&ds[dsi], &ds[nnn], len2+end_of_number2-nnn+1);
	dsi = dsi + len2+end_of_number2-nnn+1;
	si = end_of_number+1;  /* Note, there is si++ after this. */
      }
    } else {
      ds[dsi++] = s[si];  /* Copy char over */
    }
  }
  *end = slen;
  return TRUE;

}  /* decompress_string */

static void
process_str_replaced_by_name_before_return(
	   char *str_replaced_by_name,	/* IN/OUT: To be checked. */
	   int   max_str,		/* max size of str_replaced... */
	   char *crcstr,		/* Original CRCed 7 characters. */
	   int   treat_as_truncation,	/* See if we need put crc back. */
	   int   version7_signature)	/* See if prepend VERSION7 is necessary */
{
  if (treat_as_truncation && crcstr[0] != '\0' && 
      strncmp(str_replaced_by_name+strlen(str_replaced_by_name)-7, 
	      "xxxxxxx", 7) == 0)
    strcpy (str_replaced_by_name+strlen(str_replaced_by_name)-7, crcstr);

  if (version7_signature && 
      strncmp (str_replaced_by_name, VERSION7, VERSION_STRING_LENGTH) != 0)
    demangle_string_prepend (str_replaced_by_name, max_str, VERSION7);

  return;

} /* process_str_replaced_by_name_before_return */
