#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)reference_type.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991 BY                            *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      reference_type.c
**
**  ABSTRACT:
**
**	Grammar element that parses a reference type.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**  20-Feb-1998	Shi-Jung Kao
**	Clean up, _delete_type_list before return ERROR.
**
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

/* language element particular definitions */
#define ASCII_ZERO 48

int /* recognition status */
_reference_type(const char* s,		/* IN:  string to be parsed */
             int start,			/* IN:  index of first unparsed character */
             int* end,			/* OUT: index of last parsed character */
             struct MLD_reference** p)/* OUT: structure containing reference (allocated) */
{
  /* variables common to all language elements */
  int ci;  /* current index */
  char ch; /* current character */
  
  /* variables particular to this language element */
  int status;
  int end_of_type;
  struct MLD_type* t; /* storage for a type */
  struct MLD_reference* temp_p;

  /* initialization particular to language element */


  /*
  **
  ** Check for Reference Token
  **
  */

  ch = s[ci = start]; /* initialize ch and ci */
  if (ch != REFERENCE_TOKEN)
  {
    *p = 0L;
    *end == start - 1;
    return MLD_NOTHING_RECOGNIZED;
  }
  ch = s[++ci]; /* increment ch and ci */


  /*
  **
  ** Check for type
  **
  */

  status = _type(s, ci, &end_of_type, &t);
  
  switch(status)
  {
    case MLD_NOTHING_RECOGNIZED: /* there must be a type after the Reference token */
    case MLD_ERROR:
      *p = 0L;
      *end = start - 1;
      return MLD_ERROR;
    break;
    
    case MLD_SOMETHING_RECOGNIZED:
      /* normal case, continue */
    break;
  }

  /*
  **
  ** Put the reference structure together
  **
  */

  temp_p = (struct MLD_reference*) malloc(sizeof(struct MLD_reference)); 
  DEMANGLE_RECORD_ALLOC(temp_p);
  if (temp_p == 0L) {
    _delete_type_list(t);
    return MLD_ERROR; /* error, no memory allocated */
  }
  
  temp_p->type_modified = t;
  
  *p = temp_p;
  *end = end_of_type; 
#if !PRODUCTION
  if (demangle_dump)
    _dump_reference(temp_p);
#endif
  return MLD_SOMETHING_RECOGNIZED;
}
