#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)template_type.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991 BY                            *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      template_type.c
**
**  ABSTRACT:
**
**	Grammar element that parses a template_type.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**	05-Feb-1999  Shi-Jung Kao
**	Extend to ANSI mode, address type may contain qualification.
**
**	22-Apr-1999  Shi-Jung Kao
**	Ignore constant type of L10 or L_1_0 case.  i.e. return int for iL10.
**
**	01-Mar-2000  Shi-Jung Kao
**	Added DEMANGLE_TRACEIN/TRACEOUT.
**
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

int /* recognition status */
_template_type(const char* s,		/* IN:  string to be parsed */
              int start,		/* IN:  index of first unparsed character */
              int* end,			/* OUT: index of last parsed character */
              struct MLD_type** tt)	/* OUT: template type */
{
  /* variables particular to this language element */
  int status;
  int end_of_grammar_element;
  char ta[STRSIZ]; /* template argument */
  struct MLD_type* t; /* type */
  struct MLD_unmodified_type* ut; /* unmodified type */

  ta[0] = '\0';

  DEMANGLE_TRACEIN(_template_type);

/*
***
*** Get type (if it exists) -- OR --
***
*/
  
  status = _type(s, start, &end_of_grammar_element, &t);
  if (status == MLD_ERROR) {
    *end = start - 1;
    *tt = 0L;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_ERROR;
  }
  if (status == MLD_SOMETHING_RECOGNIZED)
  {
    int id;
    int ci = end_of_grammar_element;
    if (_is_special_constant_basic_type(t, &id)) {
      /* Check case of CiL_1_5 */
      status = _constant_type2(s, id, ci+1, 
			       &end_of_grammar_element, ta, STRSIZ);
      if (status == MLD_SOMETHING_RECOGNIZED) {
	if (strcmp(&ta[0], "0") != 0) { /* Replace with non-zero e.g. "5" */
	  _delete_type_structure(t);
	  status = _make_constant_type(ta, tt);
	} else /* Otherwise ignore a constant "0". */
	  *tt = t;

	*end = end_of_grammar_element;
	DEMANGLE_TRACEOUT(_template_type);
	return status;
      }
    }
    *end = end_of_grammar_element;
    *tt = t;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_SOMETHING_RECOGNIZED;
  }
  
  
/*
***
*** -- OR -- Get constant type if any exists
***
*/

  status = _constant_type(s, start, &end_of_grammar_element, ta, STRSIZ);
  if (status == MLD_ERROR) {
    *end = start - 1;
    *tt = 0L;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_ERROR;
  }
  if (status == MLD_SOMETHING_RECOGNIZED) {
    status = _make_constant_type(ta, tt);
    *end = end_of_grammar_element;
    DEMANGLE_TRACEOUT(_template_type);
    return status;
  }

  
/*
***
*** -- OR -- Get address type if any exists
***
*/
#if DEC_ANSI_FORM
  status = _address_type(s, start, &end_of_grammar_element, &t);
  if (status == MLD_ERROR)
  {
    *end = start - 1;
    *tt = 0L;
    _delete_type_list(t);
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_ERROR;
  }
  if (status == MLD_SOMETHING_RECOGNIZED)
  {
    *end = end_of_grammar_element;
    *tt = t;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_SOMETHING_RECOGNIZED;
  }
#else
  status = _address_type(s, start, &end_of_grammar_element, ta, STRSIZ);

  if (status == MLD_ERROR)
  {
    *end = start - 1;
    *tt = 0L;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_ERROR;
  }
  if (status == MLD_SOMETHING_RECOGNIZED)
  {
    /* create unmodified type */
    ut = (struct MLD_unmodified_type*) malloc(sizeof(struct MLD_unmodified_type));
    DEMANGLE_RECORD_ALLOC(ut);
    if (ut == 0L) return MLD_ERROR;  /* error, no memory allocated */
    ut->type_of_type = MLD_ADDRESS_TYPE;
    ut->the_type.type_name = (char*) malloc(strlen(ta)+1);
    ut->shared_id = 0L;
    DEMANGLE_RECORD_ALLOC(ut->the_type.type_name);
    if (ut->the_type.type_name == 0L)
    {
      DEMANGLE_FREE(ut);
      DEMANGLE_TRACEOUT(_template_type);
      return MLD_ERROR;
    }
    strcpy(ut->the_type.type_name, ta);

    /* create type structure */
    t = (struct MLD_type*) malloc(sizeof(struct MLD_type)); 
    DEMANGLE_RECORD_ALLOC(t);
    if (t == 0L)
    {
      DEMANGLE_FREE(ut->the_type.type_name);
      DEMANGLE_FREE(ut);
      DEMANGLE_TRACEOUT(_template_type);
      return MLD_ERROR; /* error, no memory allocated */
    }
    
    t->ut = ut;
    t->shared_ut = 0;
    t->tm_list = 0L; /* no type modifiers */
    t->next = 0L; /* no next, yet */
    t->repeat_type = 0L;
    t->repeat_count = 0;

    *end = end_of_grammar_element;
    *tt = t;
    DEMANGLE_TRACEOUT(_template_type);
    return MLD_SOMETHING_RECOGNIZED;
  }
#endif
  
/*
***
*** Nothing recognized, at this point
***
*/
  _delete_type_list(t);
  *end = start - 1;
  *tt = 0L;
  DEMANGLE_TRACEOUT(_template_type);
  return MLD_NOTHING_RECOGNIZED;  
}
