#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)utilities.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991, 1996 BY                      *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      utilities.c
**
**  ABSTRACT:
**
**	This module contains functions that deallocate storage.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**      06-Feb-1995	Harold Seigel
**      Add support for __aligned.
**
**      14-Feb-1995	Harold Seigel
**	Fix check for MLD_NO_SPACES in MLD_concat_mangled_template_name.
**	Put a space between '>'s for demangled nested template names.
**
**      18-Apr-1995	Harold Seigel
**	Fix problems with pointers to pointers to functions.
**
**       5-May-1995	Harold Seigel
**	Fix problems with pointers to members and member functions.
**
**      11-Aug-1995	Harold Seigel
**	Fix problems in _concat_pointers() with pointers to multi-dimensional
**	arrays.
**
**      11-Aug-1995	Harold Seigel
**	In _delete_mangled_template_name(), check for null input pointer.
**
**	20-Mar-1997	Shi-Jung Kao	cxxc_bugs: 4032, 3951
**	Add _concat_ptr_to_function.  i.e. (*(*)())	4032
**	Exclude remove_all_space on "> >".		3951
**
**	29-Oct-1997	Shi-Jung Kao
**	Add __T_ support. A typeinfo variable to describe runtime ...
**
**	20-Feb-1998	Shi-Jung Kao
**	_delete_qualifications rather than just free.  Make sure all delete
**	functions verify the non-null argument before access to it.
**
**	28-Oct-1998	Shi-Jung Kao
**	Emit Derived<int>::Derived(int) instead of 
**	Derived<int>::Derived<int>(int). Add _concat_complex_notemplatearg 
**	and MLD_concat_mangled_template_name_noarg
**
**	08-Jan-1999	Shi-Jung Kao
**	Added MLD_INIT_INTERNAL_NAME, MLD_FINI_INTERNAL_NAME,
**	MLD_PROMOTED_ENTITY_NAME.
**
**	22-Nov-1999	Shi-Jung Kao
**	Implemented shared_id for unmodified_type.
**
**	17-Jan-2000	Shi-Jung Kao
**	Implemented no_template_details. i.e. no <T0=...>
**
**	21-Jan-2000	Shi-Jung Kao
**	Supported version7, output template conversion operator
**	"__opZ1Z__tm__2_i__1CFv_Z1Z" as "T0 C::operator<T0=int> T0()"
**
**	28-Jan-2000	Shi-Jung Kao
**	Supported template variables.
**	Added _init_shared_id.  (No longer a requirement of compatible
**	with cplus-dem.c of ecc demangle on linux.)
**
**	09-Feb-2000	Shi-Jung Kao
**	Supported nontype_dim in array_type. i.e. nontype dimension.
**	Supported nested template nontype.
**	Created _alloc_nontype() to prevent every caller does it own.
**
**	14-Feb-2000     Shi-Jung Kao
**    	Created string_prepend(s,n) routines. Updated message for vtable.
**    	Display extern "C" if necessary.
**	Added DEMANGLE_TRACEIN/TRACEOUT.
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#include "dmn_str.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include "demangle_internal.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

int MLD_ias; /* inadequate array size, 0 false 1 true */

int demangle_no_template_details;

void /* substitute string or null, if nothing found */
_basic_operator_substitute(const char* ts,		/* IN:  operator token */
                          char* s,			/* OUT:	operator string */
                          int size_of_s)	/* IN:	max size of s */
{
  /* variables particular to this language element */
  char* basic_operators[NUMBER_OF_BASIC_OPERATORS + 1][2]   = BASIC_OPERATORS;
  int i; /* temporary index variable */
  char* temp;

  DEMANGLE_TRACEIN(_basic_operator_substitute);

  s[0] = 0; /* initialize to empty string */

  /*
  ***
  *** Check to see if the current character is one of the basic type characters.
  *** If so, return the corresponding string 
  */

  for (i = 0 ; i < NUMBER_OF_BASIC_OPERATORS ; i++)
  {
    if (strcmp(ts, basic_operators[i][BASIC_OPERATOR_TOKEN]) == 0) /* compare tokens */
    {
      /* copy operator string into s, if there's enough size */
      if ((9 + strlen(basic_operators[i][BASIC_OPERATOR_STRING])) < size_of_s)
      {
        strcpy(s, "operator ");
        strcat(s, basic_operators[i][BASIC_OPERATOR_STRING]);
        break;
      }
      else
      {
        if (size_of_s > 0)
        {
          strcpy(s, "");
        }
      }
    }
  }
  
  DEMANGLE_TRACEOUT(_basic_operator_substitute);
}

/*
***
*** Added sting_prepend, string_prepends, string_prependn.
*** Note, it not only returns result, also put results in "s".
*/

static char*
string_prependn(char* s, int max_s, const char* d, int n);

char*
demangle_string_prepend (char* s, 
			 int max_s, 
			 const char* d)
{
  if (d != NULL && *d != '\0')
    {
      return string_prependn(s, max_s, d, strlen(d));
    }
}

static char*
string_prepends (char* s,
		 int max_s,
		 char* d)
{
  if (strlen(s)+strlen(d) < max_s)
    {
      return string_prependn(s, max_s, d, strlen(d));
    }
}

static char*
string_prependn (s, max_s, d, n)
     char *s;
     int max_s;
     const char *d;
     int n;
{
  char *q;

  if (n > 0 && (strlen(s)+n < max_s))
    {
      for (q = s + strlen(s) + 1; q >= s; q--)
	q[n] = q[0];
      memcpy (s, d, n);
      return s;
    }
}

char*
_concat(char* s,
       int max_s,
       const char* d)
{
  if ((strlen(s)+strlen(d)) < max_s)
  {
    strcat(s,d);
  } else MLD_ias = 1;
  return s;
}

      
char*
_itostr(char* s,
       int max_s,
       int i)
{
  int t; /* temporary */
  int c; /* count */
  int si; /* string index */
  int x;
 
  /* find out how many digits i has */
  t = 1; c = 0;
  while(i/t){ t = t * 10; c++; } /* c is the number of digits */

  if (c >= max_s) return 0L; /* not enough room! */

  /* loop for each digit */
  si = 0; x = i;
  for (t = t/10 ; t > 0 ; t = t/10)
  {   
    s[si++] = (x/t) + 48;
    x = x%t;
  }
  s[si] = 0; /* null terminate */
   
  return s;
}


void
_dump_qualifications(struct MLD_qualification* ql)
{   
    struct MLD_qualification* temp_q;

    /* return all storage */
    while(ql) /* while not null */
    {
      temp_q = ql->next; /* preserve next ql */
      if ((ql->q) != 0L) {
        _dump_complex(ql->q);
      }
      ql = temp_q;
    }
}

void
_delete_qualifications(struct MLD_qualification* ql)
{   
    struct MLD_qualification* temp_q;

    DEMANGLE_TRACEIN(_delete_qualifications);

    /* return all storage */
    while(ql) /* while not null */
    {
      temp_q = ql->next; /* preserve next ql */
      
      /* delete */
      if ((ql->q) != 0L)
      {
        _delete_complex(ql->q);
      }
      DEMANGLE_FREE(ql);
      
      ql = temp_q;
    }
    DEMANGLE_TRACEOUT(_delete_qualifications);
}

void
_concat_qualifications(struct MLD_qualification* ql, 
                      char* s,
                      int max_s, 
                      int flags )
{
    struct MLD_qualification* temp_ql;

    DEMANGLE_TRACEIN(_concat_qualifications);

    /* concat all qualifications */
    while(ql) /* while not null */
    {
      temp_ql = ql->next; /* get next ql */
      
      /* concat */
      if (ql->q != 0L)
      {
        _concat_complex(ql->q, s, max_s, flags);
        if (ql->next != 0L)
        {
          _concat(s, max_s, "::"); /* add colons between qualifications */
        }
      }
      
      ql = temp_ql;
    }
    DEMANGLE_TRACEOUT(_concat_qualifications);
}

#if DEC_ANSI_FORM
void
_dump_address(struct MLD_address* ad)
{   
    _dump_qualifications(ad->qualifications);
    printf("address:%s ",ad->type_name);
}

void
_delete_address(struct MLD_address* ad)
{   
    if (ad == 0L)
      return;
    _delete_qualifications(ad->qualifications);
    DEMANGLE_FREE(ad->type_name);
    DEMANGLE_FREE(ad);
}

void
_concat_address(struct MLD_address* ad, 
                      char* s,
                      int max_s, 
                      int flags )
{
  DEMANGLE_TRACEIN(_concat_address);

  _concat(s, max_s, "&");
  if (ad->qualifications) {
    _concat_qualifications(ad->qualifications, s, max_s, flags);
    _concat(s, max_s, "::");
  }
  _concat(s, max_s, ad->type_name);
  DEMANGLE_TRACEOUT(_concat_address);
}

void
_dump_nontype(struct MLD_nontype* nt)
{   
    _dump_qualifications(nt->qualifications);
    printf("address:%s ",nt->nontype_name);
}

struct MLD_nontype*
_alloc_nontype()
{
  struct MLD_nontype* nt = 0L;
  nt = (struct MLD_nontype*) malloc(sizeof(struct MLD_nontype));
  DEMANGLE_RECORD_ALLOC(nt);
  nt->is_address = FALSE;
  nt->qualifications = 0L;
  nt->nontype_name = 0L;
  nt->nested_args = 0L;
  nt->ut = 0L;
  return nt;
}

void
_delete_nontype(struct MLD_nontype* nt)
{   
    if (nt == 0L)
      return;
    DEMANGLE_TRACEIN(_delete_nontype);

    _delete_qualifications(nt->qualifications);
    DEMANGLE_FREE(nt->nontype_name);
    _delete_mangled_template_name(nt->nested_args);
    /* Do not delete the nt->ut, its shared. */
    DEMANGLE_FREE(nt);
    DEMANGLE_TRACEOUT(_delete_nontype);
}

void
_concat_nontype(struct MLD_nontype* nt, 
                      char* s,
                      int max_s, 
                      int flags )
{
  DEMANGLE_TRACEIN(_concat_nontype);

  if (nt->is_address)
    _concat(s, max_s, "&");
  if (nt->qualifications) {
    _concat_qualifications(nt->qualifications, s, max_s, flags);
    _concat(s, max_s, "::");
  }
  if (nt->ut) {
    /* Prefer shared_id over its type. */
    if (nt->ut->shared_id && !demangle_no_template_details)
      _concat(s, max_s, nt->ut->shared_id);
    else
      _concat_unmodified_type(nt->ut, s, max_s, flags); 
  } else
    _concat(s, max_s, nt->nontype_name);
  if (nt->nested_args) {
    /* Add "<>" here, its disabled in concat_mangled_template_name. */
    _concat(s, max_s, "<");
    MLD_concat_mangled_template_name(nt->nested_args, s, max_s, flags);
    if (s[strlen(s)-1] != '>')
      _concat(s, max_s, ">");
    else
      _concat(s, max_s, " >");
  }
  DEMANGLE_TRACEOUT(_concat_nontype);    
}
#endif

void
_delete_type_modifier_list(struct MLD_type_modifiers* tml)
{   
  if (tml == 0L) return;
  DEMANGLE_TRACEIN(_delete_type_modifier_list);

  DEMANGLE_FREE(tml);
  DEMANGLE_TRACEOUT(_delete_type_modifier_list);
}

void
_concat_type_modifier_list(struct MLD_type_modifiers* tml, 
                          char* s,
                          int max_s, 
                          int flags )
{
    if (tml == 0L) return;

    DEMANGLE_TRACEIN(_concat_type_modifier_list);

    /* concat all type modifiers */
    
    if (tml->const_tm)
    {
      _concat(s, max_s, CONST_STRING);
      _concat(s, max_s, " ");
    }
    
    if (tml->signed_tm)
    {
      _concat(s, max_s, SIGNED_STRING);
      _concat(s, max_s, " ");
    }
    
    if (tml->unaligned_tm)
    {
      _concat(s, max_s, UNALIGNED_STRING);
      _concat(s, max_s, " ");
    }

    if (tml->unsigned_tm)
    {
      _concat(s, max_s, UNSIGNED_STRING);
      _concat(s, max_s, " ");
    }
    
    if (tml->volatile_tm)
    {
      _concat(s, max_s, VOLATILE_STRING);
      _concat(s, max_s, " ");
    }    
    DEMANGLE_TRACEOUT(_concat_type_modifier_list);
}

void
_dump_type_modifier_list(struct MLD_type_modifiers* tml)
{
    if (tml == 0L) return;

    /* dump all type modifiers */

    if (tml->const_tm)
      printf("%s ", CONST_STRING);
    
    if (tml->signed_tm)
      printf("%s ", SIGNED_STRING);
    
    if (tml->unaligned_tm)
      printf("%s ", UNALIGNED_STRING);

    if (tml->unsigned_tm)
      printf("%s ", UNSIGNED_STRING);
    
    if (tml->volatile_tm)
      printf("%s ", VOLATILE_STRING);
}

void
_dump_basic(int bt)
{
    char* basic_type_strings[NUMBER_OF_BASIC_TYPES + 
			    NUMBER_OF_BASIC3_TYPES + 1] = BASIC_TYPE_STRINGS;

    if ((bt >= -1) && (bt <= NUMBER_OF_BASIC_TYPES))
      printf("%s ",basic_type_strings[bt]);
}

void 
_dump_unmodified_type(struct MLD_unmodified_type* ut)
{
  if (ut == 0L) return;
  
  switch(ut->type_of_type) {
  case MLD_ADDRESS_TYPE:
#if DEC_ANSI_FORM
    printf("address:");
    _dump_address(ut->the_type.address);
#else
    printf("address:%s ",ut->the_type.type_name);
#endif
    break;
  case MLD_CONSTANT_TYPE:
    printf("constant:%s ",ut->the_type.type_name);
    break;

  case MLD_COMPLEX_TYPE:
    printf("complex:");
    _dump_qualifications(ut->the_type.complex);
    break;
        
  case MLD_BASIC_TYPE:
    printf("basic:");
    _dump_basic(ut->the_type.basic);
    break;
        
  case MLD_POINTER_TYPE:
    _dump_pointer(ut->the_type.pointer);
    break;

  case MLD_NONTYPE_TYPE:
    _dump_nontype(ut->the_type.nontype);
    break;

  case MLD_REFERENCE_TYPE:
    printf("reference:");
    _dump_reference(ut->the_type.reference);
    break;
        
  case MLD_ARRAY_TYPE:
    printf("array:");
    _dump_array(ut->the_type.array);
    break;
        
  case MLD_FUNCTION_TYPE:
    _dump_function(ut->the_type.function);
    break;
        
  case MLD_POINTER_TO_MEMBER_TYPE:
    printf("pointer_to_member:");
    _dump_pointer_to_member(ut->the_type.ptr_to_member);
    break;  
  }
}

void 
_delete_unmodified_type(struct MLD_unmodified_type* ut)
{
  if (ut == 0L) return;
  DEMANGLE_TRACEIN(_delete_unmodified_type);

  if (ut->shared_id)
    DEMANGLE_FREE(ut->shared_id);
  switch(ut->type_of_type) 
  {
        case MLD_ADDRESS_TYPE:
#if DEC_ANSI_FORM
	  _delete_address(ut->the_type.address);
	  break;
#endif
        case MLD_CONSTANT_TYPE:
	  DEMANGLE_FREE(ut->the_type.type_name);
        break;

        case MLD_COMPLEX_TYPE:
	  _delete_qualifications(ut->the_type.complex);
	  break;
        
        case MLD_BASIC_TYPE:
          /* do nothing, basic type is an int */
        break;
        
        case MLD_POINTER_TYPE:
          _delete_pointer(ut->the_type.pointer);
        break;
        
        case MLD_NONTYPE_TYPE:
          _delete_nontype(ut->the_type.nontype);
        break;
        
        case MLD_REFERENCE_TYPE:
          _delete_reference(ut->the_type.reference);
        break;
        
        case MLD_ARRAY_TYPE:
          _delete_array(ut->the_type.array);
        break;
        
        case MLD_FUNCTION_TYPE:
          _delete_function(ut->the_type.function);
        break;
        
        case MLD_POINTER_TO_MEMBER_TYPE:
          _delete_pointer_to_member(ut->the_type.ptr_to_member);
        break;  
  }
  DEMANGLE_FREE(ut);
  DEMANGLE_TRACEOUT(_delete_unmodified_type);
}

void
_concat_unmodified_type(struct MLD_unmodified_type* ut,
                      char* s,
                      int max_s, 
                      int flags)
{
  if (ut == 0L) return;
 
  DEMANGLE_TRACEIN(_concat_unmodified_type);

  /* Ignore the shared_id, concat the real one at this level. */

  switch(ut->type_of_type) 
  {
  case MLD_ADDRESS_TYPE:
#if DEC_ANSI_FORM
    _concat_address(ut->the_type.address, s, max_s, flags);
    break;
#endif
  case MLD_CONSTANT_TYPE: {
    int is_char = (strlen(ut->the_type.type_name) == 1 &&
		   (ut->the_type.type_name[0] < '0' ||
		    ut->the_type.type_name[0] > '9'));
    if (is_char) _concat(s, max_s, "\'");
    _concat(s, max_s, ut->the_type.type_name);
    if (is_char) _concat(s, max_s, "\'");
    break;
  }
  case MLD_COMPLEX_TYPE:
    _concat_qualifications(ut->the_type.complex, s, max_s, flags);
    break;
        
  case MLD_BASIC_TYPE:
    _concat_basic(ut->the_type.basic, s, max_s, flags);
    break;
        
  case MLD_POINTER_TYPE:
    _concat_pointer(ut->the_type.pointer, s, max_s, flags);
    break;
        
  case MLD_NONTYPE_TYPE:
    _concat_nontype(ut->the_type.nontype, s, max_s, flags);
    break;
        
  case MLD_REFERENCE_TYPE:
    _concat_reference(ut->the_type.reference, s, max_s, flags);
    break;
        
  case MLD_ARRAY_TYPE:
    _concat_array(ut->the_type.array, s, max_s, flags);
    break;
        
  case MLD_FUNCTION_TYPE:
    _concat_function(ut->the_type.function, s, max_s, flags);
    break;
        
  case MLD_POINTER_TO_MEMBER_TYPE:
    _concat_pointer_to_member(ut->the_type.ptr_to_member, s, max_s, flags);
    break;  
  }
  DEMANGLE_TRACEOUT(_concat_unmodified_type);
}


struct MLD_type* /* returns the "next" structure */
_dump_type_structure(struct MLD_type* t)
{
  if (t == 0L)
    return 0L;

  printf("\ntype_modifier_list:");
  _dump_type_modifier_list(t->tm_list);
  printf("\nunmodified_type:%d",t->shared_ut);
  _dump_unmodified_type(t->ut);  

  return t->next;
}

struct MLD_type* /* returns the "next" structure */
_delete_type_structure(struct MLD_type* t)
{
  struct MLD_type* t_temp;
  
  if (t == 0L)
  {
    return 0L;
  }
  
  DEMANGLE_TRACEIN(_delete_type_structure);

  _delete_type_modifier_list(t->tm_list);
  if (!(t->shared_ut))
    _delete_unmodified_type(t->ut);  
  
  t_temp = t->next;
  DEMANGLE_FREE(t);
  DEMANGLE_TRACEOUT(_delete_type_structure);
  return t_temp;
}

static char sharedCHAR = 'T';
static char sharedINT = '0';

void _init_shared_id() {
  sharedCHAR = 'T';
  sharedINT = '0';
  return;
}

void _set_shared_id(struct MLD_unmodified_type *ut) {
  DEMANGLE_TRACEIN(_set_shared_id);

  if (demangle_no_template_details || ut->shared_id)
    return;
  ut->shared_id = (char*) malloc(3);
  DEMANGLE_RECORD_ALLOC(ut->shared_id);
  if (ut->shared_id != 0L) {
    ut->shared_id[0] = sharedCHAR;
    ut->shared_id[1] = sharedINT++;
    ut->shared_id[2] = '\0';
    if (sharedINT > '9') {
      sharedINT = '0';
      sharedCHAR++;
      if (sharedCHAR > 'V')
	sharedCHAR = 'T';
    }
  }
  DEMANGLE_TRACEOUT(_set_shared_id);
}

struct MLD_type* /* returns the "next" structure */
_concat_type_structure(struct MLD_type* t,
                      char* s,
                      int max_s, 
                      int flags)
{
  if (t == 0L)
  {
    return 0L;
  }
  
  DEMANGLE_TRACEIN(_concat_type_structure);

  _concat_type_modifier_list(t->tm_list, s, max_s, flags);
  if (t->shared_ut) {
    if (t->ut && t->ut->shared_id)
      _concat(s, max_s, t->ut->shared_id);
    else if (demangle_no_template_details)
      _concat_unmodified_type(t->ut, s, max_s, flags); 
    else
      _concat(s, max_s, "ER_nosharedid");
  } else {
    if (t->ut && t->ut->shared_id){
      _concat(s, max_s, t->ut->shared_id);
      _concat(s, max_s, "=");
    }
    _concat_unmodified_type(t->ut, s, max_s, flags); 
  }
  DEMANGLE_TRACEOUT(_concat_type_structure);
  return t->next;
}

void
_dump_type_list(struct MLD_type* tl)
{
  while (tl != 0L) /* keep deleting until terminal is reached */
  {
    tl = _dump_type_structure(tl); /* returns the next type node in list */
  }
}

void
_delete_type_list(struct MLD_type* tl)
{
  DEMANGLE_TRACEIN(_delete_type_list);

  while (tl != 0L) /* keep deleting until terminal is reached */
  {
    tl = _delete_type_structure(tl); /* returns the next type node in list */
  }
  DEMANGLE_TRACEOUT(_delete_type_list);
}

void
_concat_type_list(struct MLD_type* tl,
                 char* s,
                 int max_s, 
                 int flags)
{
  DEMANGLE_TRACEIN(_concat_type_list);

  while (tl != 0L) /* keep concating till terminal is reached */
  {
#if DEC_ANSI_FORM
    int count = tl->repeat_count;

    if (count > 0) { /* Nmn */
      int i = 0;
      for (i = 0; i < count; i++) {
	_concat_type_structure(tl->repeat_type, s, max_s, flags);
	if (i < count-1)
	  _concat(s, max_s, ", ");
      }
      tl = tl->next;
    } else if (tl->repeat_type) { /* Tn */
      struct MLD_type* dummy = 
	_concat_type_structure(tl->repeat_type, s, max_s, flags);
      tl = tl->next;
    } else
#endif
    /* returns the next type node in list */
    tl = _concat_type_structure(tl, s, max_s, flags); 
    /* if this isn't the end of the type list, put in a comma */
    if ((tl != 0L) && ((strlen(s)+2) < max_s))
    {
      _concat(s, max_s, ", ");
    } 
  }
  DEMANGLE_TRACEOUT(_concat_type_list);
}


void
_dump_mangled_name(struct MLD_mangled_name* mn)
{
  if (mn == 0L)
  {
    return;
  }
  
  switch(mn->mangled_name_type) {
  case MLD_FUNCTION_MANGLED_NAME:
    _dump_mangled_function_name(mn->name.funct_name);
    break;
        
  case MLD_TEMPLATE_MANGLED_NAME:
    _dump_mangled_template_name(mn->name.templ_name);
    break;
        
  case MLD_VARIABLE_MANGLED_NAME:
    _dump_mangled_variable_name(mn->name.var_name);
    break;

  case MLD_PROMOTED_ENTITY_NAME:
    _dump_mangled_function_name(mn->name.funct_name);
    break;
        
  case MLD_RUNTIMETYPE_INFO_NAME:
  case MLD_VIRTUAL_TABLE_NAME:
  case MLD_INIT_INTERNAL_NAME:
  case MLD_FINI_INTERNAL_NAME:
    if (mn->name.info != 0L) {
      if (mn->mangled_name_type == MLD_VIRTUAL_TABLE_NAME)
	printf("vtbl ");
      else if (mn->mangled_name_type == MLD_RUNTIMETYPE_INFO_NAME)
	printf("runtime typeinfo ");
      else if (mn->mangled_name_type == MLD_INIT_INTERNAL_NAME)
	printf("internal initialization routine ");
      else if (mn->mangled_name_type == MLD_FINI_INTERNAL_NAME)
	printf("internal termination routine ");
    }
    break;
        
  case MLD_NO_MANGLED_NAME:
  default:
    /* do nothing */
    break;  
  }
}

void
MLD_delete_mangled_name(struct MLD_mangled_name* mn)
{
  if (mn == 0L)
  {
    return;
  }
  
  DEMANGLE_TRACEIN(MLD_delete_mangled_name);

  switch(mn->mangled_name_type) 
  {
        case MLD_FUNCTION_MANGLED_NAME:
        case MLD_PROMOTED_ENTITY_NAME:
          _delete_mangled_function_name(mn->name.funct_name);
        break;
        
        case MLD_TEMPLATE_MANGLED_NAME:
          _delete_mangled_template_name(mn->name.templ_name);
        break;
        
        case MLD_VARIABLE_MANGLED_NAME:
          _delete_mangled_variable_name(mn->name.var_name);
        break;
        
        case MLD_VIRTUAL_TABLE_NAME:
        case MLD_RUNTIMETYPE_INFO_NAME:
        case MLD_INIT_INTERNAL_NAME:
        case MLD_FINI_INTERNAL_NAME:
          if (mn->name.info != 0L)
          {
	    DEMANGLE_FREE(mn->name.info);
          }
        break;

        case MLD_NO_MANGLED_NAME:
        default:
          /* do nothing */
        break;  
  }
        
  /* free memory occupied by this demangled name */
  DEMANGLE_FREE(mn);
  DEMANGLE_TRACEOUT(MLD_delete_mangled_name);
}

#define NO_FLAGS 0

void
MLD_concat_mangled_name(struct MLD_mangled_name* mn,
                    char* s,
                    int max_s, 
                    int flags)
{
  /* these declarations are for checking to see if the virtual table name 
   * is actually a template name, in which case, conversion is needed
   */
  int end;
  char bs[BUFSIZ];
  char rs[BUFSIZ];
  char as[BUFSIZ];
  struct MLD_mangled_template_name* mtn;
  int status;

  if (mn == 0L) {
    return;
  }
  
  DEMANGLE_TRACEIN(MLD_concat_mangled_name);

  switch(mn->mangled_name_type) {
  case MLD_FUNCTION_MANGLED_NAME:
  case MLD_PROMOTED_ENTITY_NAME:
    _concat_mangled_function_name(mn->name.funct_name, s, max_s, flags);
    break;
        
  case MLD_TEMPLATE_MANGLED_NAME:
    MLD_concat_mangled_template_name(mn->name.templ_name, s, max_s, flags);
    break;
        
  case MLD_VARIABLE_MANGLED_NAME:
    _concat_mangled_variable_name(mn->name.var_name, s, max_s, flags);
    break;
        
  case MLD_RUNTIMETYPE_INFO_NAME:
  case MLD_VIRTUAL_TABLE_NAME:
    if ((mn->name.info != 0L) && (flags & 2)) { /* check for info flag */
      if (mn->mangled_name_type == MLD_VIRTUAL_TABLE_NAME) {
	demangle_string_prepend(s, max_s, "The virtual table '");
	_concat(s, max_s, "' for class '");
      } else if (mn->mangled_name_type == MLD_RUNTIMETYPE_INFO_NAME)
	_concat(s, max_s, "  A typeinfo variable to describe runtime '"); 
      /* mn.name.info might be a template name.  If it is, print 
	 the demangled version. */
      status = _mangled_template_name(mn->name.info, 0, &end, &mtn, bs, rs, as, BUFSIZ);

      switch (status) {
      case MLD_NOTHING_RECOGNIZED:
	_concat(s, max_s, mn->name.info); 
	break;

      case MLD_ERROR:
	_concat(s, max_s, mn->name.info); 
	break;

      case MLD_SOMETHING_RECOGNIZED:
	MLD_concat_mangled_template_name(mtn, s, max_s, NO_FLAGS);
	_delete_mangled_template_name(mtn);
	break;
      }
      if (mn->mangled_name_type == MLD_VIRTUAL_TABLE_NAME) {
	_concat(s, max_s, "' is undefined, make sure that the first noninlined virtual member function is defined.  "); 
      } else if (mn->mangled_name_type == MLD_RUNTIMETYPE_INFO_NAME)
	_concat(s, max_s, "' type information.  "); 
    }
    break;   
        
  case MLD_INIT_INTERNAL_NAME:
    if ((mn->name.info != 0L) && (flags & 2)) { /* check for info flag */
      _concat(s, max_s, "Report a problem, if found in unresolved list.\
  An internal initialization routine for file '"); 
      _concat(s, max_s, mn->name.info); 
      _concat(s, max_s, "' \n"); 
    }
    break;

  case MLD_FINI_INTERNAL_NAME:
    if ((mn->name.info != 0L) && (flags & 2)) { /* check for info flag */
      _concat(s, max_s, "Report a problem, if found in unresolved list.\
  An internal termination routine for file '"); 
      _concat(s, max_s, mn->name.info); 
      _concat(s, max_s, "' \n"); 
    }
    break;

  case MLD_NO_MANGLED_NAME:
  default:
    /* do nothing */
    break;  
  }
  DEMANGLE_TRACEOUT(MLD_concat_mangled_name);
}


void
_dump_mangled_function_name(struct MLD_mangled_function_name* mfn)
{
  /* free identifier */
  switch (mfn->function_name_type) {
  case MLD_CONVERSION_OPERATOR_FUNCTION_NAME:
    if (mfn->id.conversion_operator_type != 0L)
      _dump_type_list(mfn->id.conversion_operator_type);
    break;
        
  default:
    if (mfn->id.identifier != 0L)
      printf("%s ",mfn->id.identifier);
    break;
  }
  
  /* dump return type */
  if (mfn->rtn != 0L)
    _dump_type_structure(mfn->rtn);
  
  /* dump qualifiers */
  if (mfn->qualifications != 0L)
    _dump_qualifications(mfn->qualifications);
  
  /* dump args */
  _dump_type_list(mfn->args);
  
  /* dump function type modifiers */
  _dump_type_modifier_list(mfn->tm_list);
}

void
_delete_mangled_function_name(struct MLD_mangled_function_name* mfn)
{
  if (mfn == 0L) return;

  DEMANGLE_TRACEIN(_delete_mangled_function_name);

  /* free identifier */
  switch (mfn->function_name_type)
  {
    case MLD_CONVERSION_OPERATOR_FUNCTION_NAME:
     /* free converstion operator identifier, which is a type */
     if (mfn->id.conversion_operator_type != 0L)
     {
       _delete_type_list(mfn->id.conversion_operator_type);
     }
    break;
        
    default:
     /* free identifier */
     if (mfn->id.identifier != 0L)
     {
       DEMANGLE_FREE(mfn->id.identifier);
     }
    break;
  }
  
  /* free qualifiers */
  if (mfn->qualifications != 0L)
  {
    _delete_qualifications(mfn->qualifications);
  }

  /* free return type */
  if (mfn->rtn != 0L) {
    _delete_type_structure(mfn->rtn);
  }
 
  /* free args */
  _delete_type_list(mfn->args);
  
  /* free function type modifiers */
  _delete_type_modifier_list(mfn->tm_list);
  
  /* free memory occupied by this demangled name */
  DEMANGLE_FREE(mfn);
  DEMANGLE_TRACEOUT(_delete_mangled_function_name);
}

void
_concat_mangled_function_name(struct MLD_mangled_function_name* mfn,
                    char* s,
                    int max_s, 
                    int flags)
{
  struct MLD_complex* c;
  int status;
  char string[STRSIZ];
  int template_function = 0;
  char* fname = 0L;

  /* Under ANSI, a template_function has the form of C::foo<...>()
   * Both "C" and "foo<...>" are constructed in its qualifications.
   * Hence, template_function = qualifications.ID == mfn.ID
   * We shall skip the "::" and the ID in mfn all togather.
   */

  if (mfn == 0L) return; /* return if there isn't anything to concat */

  DEMANGLE_TRACEIN(_concat_mangled_function_name);

  /* concat return type */
  /* Under version7, ignore return type of template converson operator. */
  if (mfn->rtn != 0L && 
      mfn->function_name_type != MLD_CONVERSION_OPERATOR_FUNCTION_NAME) {
    _concat_type_structure(mfn->rtn, s, max_s, flags); 
    _concat(s, max_s, " ");
  }

  if (mfn->function_name_type != MLD_CONVERSION_OPERATOR_FUNCTION_NAME)
    fname = mfn->id.identifier;

  /* concat qualifiers */
  if (mfn->qualifications != 0L) {
    struct MLD_qualification *lastq;
    struct MLD_qualification *q;

    _concat_qualifications(mfn->qualifications, s, max_s, flags);
    
    for(q = mfn->qualifications; q; q = q->next)
      lastq = q;

    if (lastq->q->mangled_template_name && fname)
      template_function = strcmp(fname,
		 lastq->q->name.template_name->identifier) == 0;
    else if (lastq->q->mangled_template_name &&
	     mfn->function_name_type == MLD_CONVERSION_OPERATOR_FUNCTION_NAME) {
      /* Under version7, output "T0 C::operator<T0=int> T0()", this
       * is compatible to existing "C::operaotr int()".
       */
      template_function = TRUE;
    }

    if (!template_function) {
      /* add "::" after the qualifications */
      _concat(s, max_s, "::");
    }
  }

  /* concat identifier */
  switch (mfn->function_name_type) {
  case MLD_CONVERSION_OPERATOR_FUNCTION_NAME:
    if (mfn->id.conversion_operator_type != 0L)
      {
	/* add keyword operator before conversion op */
	if (template_function) /* Not for template conversion operator. */
	  _concat(s, max_s, " ");
	else
	  _concat(s, max_s, "operator ");
	_concat_type_list(mfn->id.conversion_operator_type, s, max_s, flags);
      }
    break;
    
  case MLD_DESTRUCTOR_FUNCTION_NAME: 
    /* concatenate a tilde */
    _concat(s, max_s, "~");
    /* note:  the lack of break is intentional */
      
  case MLD_CONSTRUCTOR_FUNCTION_NAME:
    /* get the class name */
    status = _get_class( mfn->qualifications, &c );
		  
    /* concatenate the class name */
    _concat_complex_notemplatearg(c, s, max_s, flags);
	
    break;

    
  case MLD_NO_FUNCTION_NAME: /* don't do anything */
    break;
    
  case MLD_OPERATOR_FUNCTION_NAME:
    _basic_operator_substitute(mfn->id.identifier, string, STRSIZ);
    if (string != 0L)
      {
	_concat(s, max_s, string);
      }      
    break;
    
  case MLD_NORMAL_FUNCTION_NAME:
  default:
    if (!template_function && mfn->id.identifier != 0L)
      {
	_concat(s, max_s, mfn->id.identifier);
      }
    break;
  }
  
  /* concat args */
  if (mfn->args != 0L)
  {  
    _concat(s, max_s, "("); /* begin type list with paren */
  }
  
  _concat_type_list(mfn->args, s, max_s, flags);
  
  if (mfn->args != 0L)
  {  
    _concat(s, max_s, ")"); /* begin type list with paren */
  }
  
  /* concat type modifiers */
  if (mfn->tm_list != 0L)
  {
    _concat(s, max_s, " ");
    _concat_type_modifier_list(mfn->tm_list, s, max_s, flags);
  }
  DEMANGLE_TRACEOUT(_concat_mangled_function_name);
}                    

void
MLD_concat_mangled_function_name_identifier(struct MLD_mangled_function_name* mfn,
                                        char* s,
                                        int max_s,
                                        int flags)
{
  struct MLD_complex* c;
  int status;
  char string[STRSIZ];

  if (mfn == 0L) return; /* return if there isn't anything to concat */

  DEMANGLE_TRACEIN(MLD_concat_mangled_function_name_identifier);

  /* concat identifier */
  switch (mfn->function_name_type)
  {
    case MLD_CONVERSION_OPERATOR_FUNCTION_NAME:
     if (mfn->id.conversion_operator_type != 0L)
     {
       _concat(s, max_s, "operator "); /* add keyword operator before conversion op */
       _concat_type_list(mfn->id.conversion_operator_type, s, max_s, flags);
     }
    break;

    case MLD_DESTRUCTOR_FUNCTION_NAME:
      /* concatenate a tilde */
      _concat(s, max_s, "~");
      /* note:  the lack of break is intentional */

    case MLD_CONSTRUCTOR_FUNCTION_NAME:
      /* get the class name */
      status = _get_class( mfn->qualifications, &c );

      /* concatenate the class name */
      _concat_complex_notemplatearg(c, s, max_s, flags);
    break;

    case MLD_NO_FUNCTION_NAME: /* don't do anything */
    break;

    case MLD_OPERATOR_FUNCTION_NAME:
     _basic_operator_substitute(mfn->id.identifier, string, STRSIZ);
     if (string != 0L)
     {
       _concat(s, max_s, string);
     }
    break;

    case MLD_NORMAL_FUNCTION_NAME:
    default:
     if (mfn->id.identifier != 0L)
     {
       _concat(s, max_s, mfn->id.identifier);
     }
    break;
  }
  DEMANGLE_TRACEOUT(MLD_concat_mangled_function_name_identifier);
}


void
_dump_mangled_template_name(struct MLD_mangled_template_name* mtn)
{

  if (mtn == 0L) return;

  /* dump identifier */
  if (mtn->identifier != 0L)
    printf("%s ",mtn->identifier);
  
  /* dump args */
  _dump_type_list(mtn->args);
}

void
_delete_mangled_template_name(struct MLD_mangled_template_name* mtn)
{

  if (mtn == 0L) return;

  DEMANGLE_TRACEIN(_delete_mangled_template_name);

  /* free identifier */
  if (mtn->identifier != 0L)
  {
    DEMANGLE_FREE(mtn->identifier);
  }  
  
  /* free args */
  _delete_type_list(mtn->args);
  
  /* free memory occupied by this demangled name */
  DEMANGLE_FREE(mtn);
  DEMANGLE_TRACEOUT(_delete_mangled_template_name);
}

void
_remove_all_spaces(char* s)
{
  int ci, bi; /* current index, bubble index */
  int length_of_s;
  char ch; 
  
  DEMANGLE_TRACEIN(_remove_all_spaces);

  length_of_s = strlen(s);
  for (ci = 0, ch = s[ci] ; ci < length_of_s ; ci++, ch = s[ci]) /* check each character */
  {
    if (ch == ' ') /* current character is a space, bubble other characters up */
    {
      if (ci > 0 && s[ci-1] == '>' && ci < length_of_s && s[ci+1] == '>')
	continue;  /* Leave "> >" alone, cxxc_bugs 3951 */

      /* "bubble" other characters up */
      for (bi = ci ; bi < length_of_s ; bi++)
      {
        s[bi] = s[bi+1];
      }

      /* decrease length of s by one */
      length_of_s--;
    } 
  }
  DEMANGLE_TRACEOUT(_remove_all_spaces);
}

void
MLD_concat_mangled_template_name(struct MLD_mangled_template_name* mtn,
                    char* s,
                    int max_s, 
                    int flags)
{
  char* id;
  char temp_s[STRSIZ]; temp_s[0] = 0; /* temp_s = "" */

  if (mtn == 0L) return; /* return if there isn't anything to concat */

  DEMANGLE_TRACEIN(MLD_concat_mangled_template_name);

  /* concat identifier */
  id = mtn->identifier;
  if (id != 0L)
  {
    /* Output the "__opZ1Z" template conversion operator as "operator" */
    if (strncmp (mtn->identifier, "__opZ", 5) == 0)
      _concat(temp_s, STRSIZ, "operator");
    else
      _concat(temp_s, STRSIZ, mtn->identifier);
  }
  
  if (mtn->args != 0L && id != 0L)
  {
    /* Without identifier, its an arg, so skip the "<". */
    _concat(temp_s, STRSIZ, "<"); /* begin type list with < */
  }
  
  /* concat args */
  _concat_type_list(mtn->args, temp_s, STRSIZ, flags);
  
  if (mtn->args != 0L && id != 0L)
  {  
    if (temp_s[strlen(temp_s)-1] != '>')
    	_concat(temp_s, STRSIZ, ">"); /* end type list with '>' */
    else
	/*
	 * If a nested template instantiation, put '> >', not '>>' in
	 * the demangled string.  The blank is beneficial because the
	 * demangled name can then be used in a #pragma define_template
	 * statement.  Names with '>>' cannot be used because the compiler
	 * thinks that '>>' is a shift operator.
	 */
    	_concat(temp_s, STRSIZ, " >"); /* end type list with ' >' */
  }

  /* remove all spaces if flag is set */
  if (flags & MLD_NO_SPACES)
  {
    _remove_all_spaces(temp_s);
  }

  _concat(s, max_s, temp_s);
  DEMANGLE_TRACEOUT(MLD_concat_mangled_template_name);
}

void
MLD_concat_mangled_template_name_noarg(struct MLD_mangled_template_name* mtn,
                    char* s,
                    int max_s, 
                    int flags)
{
  char temp_s[STRSIZ]; temp_s[0] = 0; /* temp_s = "" */

  if (mtn == 0L) return; /* return if there isn't anything to concat */

  DEMANGLE_TRACEIN(MLD_concat_mangled_template_name_noarg);

  /* concat identifier */
  if (mtn->identifier != 0L)
    _concat(s, max_s, mtn->identifier);
  DEMANGLE_TRACEOUT(MLD_concat_mangled_template_name_noarg);
}

void
_dump_mangled_variable_name(struct MLD_mangled_variable_name* mvn)
{
  if (mvn == 0L) return;  /* no need to do anything */

  if (mvn->identifier != 0L)
    printf("%s ",mvn->identifier);

  if (mvn->qualifications)
    _dump_qualifications(mvn->qualifications);
}

void
_delete_mangled_variable_name(struct MLD_mangled_variable_name* mvn)
{
  if (mvn == 0L) return;  /* no need to do anything */

  DEMANGLE_TRACEIN(_delete_mangled_variable_name);

  if (mvn->identifier != 0L)
  {
    DEMANGLE_FREE(mvn->identifier);
  }

  if (mvn->qualifications)
  {
    _delete_qualifications(mvn->qualifications);
  }
  DEMANGLE_FREE(mvn);
  DEMANGLE_TRACEOUT(_delete_mangled_variable_name);
}

void
_concat_mangled_variable_name(struct MLD_mangled_variable_name* mvn,
                             char* s,
                             int max_s, 
                             int flags)
{
  int skip_identifier = 0;

  if (mvn == 0L) return; /* return if there isn't anything to concat */

  DEMANGLE_TRACEIN(_concat_mangled_variable_name);

  /* concat qualifiers */
  if (mvn->qualifications != 0L)
  {
    struct MLD_qualification *lastq;
    struct MLD_qualification *q;

    _concat_qualifications(mvn->qualifications, s, max_s, flags);

    for(q = mvn->qualifications; q; q = q->next)
      lastq = q;

    if (lastq->q->mangled_template_name && mvn->identifier &&
        strcmp(mvn->identifier,
	       lastq->q->name.template_name->identifier) == 0) {
      /* A template variable, skip the ::mvn->identifier */
      skip_identifier = 1;

    } else
      /* add "::" after the qualifications */
      _concat(s, max_s, "::");    
  }

  /* concat identifier */
  if (mvn->identifier != 0L && !skip_identifier)
  {
    _concat(s, max_s, mvn->identifier);
  }  
  DEMANGLE_TRACEOUT(_concat_mangled_variable_name);
}                 

void
_dump_complex(struct MLD_complex* c)
{
  if (c == 0L) return;

  /* dump the name depending on what type of name it is */
  if ( c->mangled_template_name )
    _dump_mangled_template_name(c->name.template_name);
  else if (c->name.structure_name)
    printf("%s ",c->name.structure_name);
}

void
_delete_complex(struct MLD_complex* c)
{
  if (c == 0L) return;

  DEMANGLE_TRACEIN(_delete_complex);

  /* delete the name depending on what type of name it is */
  if ( c->mangled_template_name )
  {
    _delete_mangled_template_name(c->name.template_name);
  }
  else
  {
    if (c->name.structure_name) {
      DEMANGLE_FREE(c->name.structure_name);
    }
  }
  
  /* delete self */
  DEMANGLE_FREE(c);
  DEMANGLE_TRACEOUT(_delete_complex);
}

void
_concat_complex(struct MLD_complex* c,
               char* s,
               int max_s, 
               int flags)
{
  if (c == 0L) return;

  DEMANGLE_TRACEIN(_concat_complex);

  /* concat the name depending on what type of name it is */
  if (c->mangled_template_name)
  {
    MLD_concat_mangled_template_name(c->name.template_name, s, max_s, flags); 
  }
  else
  {
    _concat(s, max_s, c->name.structure_name);
  }  
  DEMANGLE_TRACEOUT(_concat_complex);
}


void
_concat_complex_notemplatearg(struct MLD_complex* c,
               char* s,
               int max_s, 
               int flags)
{
  if (c == 0L) return;

  DEMANGLE_TRACEIN(_concat_complex_notemplatearg);

  /* concat the name depending on what type of name it is */
  if (c->mangled_template_name)
    MLD_concat_mangled_template_name_noarg(
	   c->name.template_name, s, max_s, flags); 
  else
    _concat(s, max_s, c->name.structure_name);
  DEMANGLE_TRACEOUT(_concat_complex_notemplatearg);
}


void
_concat_basic(int bt,
               char* s,
               int max_s, 
               int flags)
{
    char* basic_type_strings[NUMBER_OF_BASIC_TYPES +
			    NUMBER_OF_BASIC3_TYPES + 1] = BASIC_TYPE_STRINGS;
    DEMANGLE_TRACEIN(_concat_basic);

    if ((bt >= -1) && (bt <= NUMBER_OF_BASIC_TYPES+NUMBER_OF_BASIC3_TYPES))
    {
      _concat(s, max_s, basic_type_strings[bt]);
    }
    DEMANGLE_TRACEOUT(_concat_basic);
}


void
_dump_pointer(struct MLD_pointer* p)
{
  if (p == 0L) return;

  /* dump type modified */
  printf("(PTR:");
  _dump_type_structure(p->type_modified);
  printf(":PTR)\n");
}


void
_delete_pointer(struct MLD_pointer* p)
{
  if (p == 0L) return;

  DEMANGLE_TRACEIN(_delete_pointer);

  /* delete type modified */
  _delete_type_structure(p->type_modified);
  
  /* delete self */
  DEMANGLE_FREE(p);
  DEMANGLE_TRACEOUT(_delete_pointer);
}


_concat_ptr_to_func_ptrs(struct MLD_pointer* p,
               char* s,
               int max_s,
               int flags)
{
  DEMANGLE_TRACEIN(_concat_ptr_to_func_ptrs);

    if (p->type_modified->ut->type_of_type != MLD_FUNCTION_TYPE)
	_concat_ptr_to_func_ptrs (p->type_modified->ut->the_type.pointer,
		s, max_s, flags);

    _concat (s, max_s, "*");
    _concat_type_modifier_list(p->type_modified->tm_list, s, max_s, flags);
  DEMANGLE_TRACEOUT(_concat_ptr_to_func_ptrs);
}

#define MAX_T 4096

static void
_concat_ptr_to_function(struct MLD_function* p,
               char* s,
               int max_s, 
	       char *t,
	       int max_t,
               int flags)
{
  /* ptr_to_function_type was built at function_type */

  DEMANGLE_TRACEIN(_concat_ptr_to_function);

  _concat(s, max_s, "(*");
  _concat(t, max_t, ")(");
  _concat_type_list(p->args, t, max_t, flags);
  _concat(t, max_t, ")");

  if (p->ptr_to_function_type)
    _concat_ptr_to_function(p->ptr_to_function_type, s, max_s, t, max_t, flags);
  DEMANGLE_TRACEOUT(_concat_ptr_to_function);
}

void
_concat_pointer(struct MLD_pointer* p,
               char* s,
               int max_s, 
               int flags)
{
  if (p == 0L) return;

  DEMANGLE_TRACEIN(_concat_pointer);

  
  /* if this is a pointer to a function, print in a special way... */
  if (p->type_modified->ut->type_of_type == MLD_FUNCTION_TYPE)
  {
    struct MLD_function* func = p->type_modified->ut->the_type.function;
    char t[MAX_T] ="";

    /* Display if is extern "C". */
    if (func && func->extern_c)
      _concat (s, max_s, "extern \"C\" ");

    /* concatenate return type */
    _concat_type_structure(func->return_type, s, max_s, flags); 
    
    /* concatenate "( <type_modifiers> *)" */
    if (p->type_modified->tm_list) {
      _concat(s, max_s, " (");
      _concat_type_modifier_list(p->type_modified->tm_list, s, max_s, flags);
      _concat(s, max_s, "*)(");
    } else {
      if (func->ptr_to_function_type) {
	_concat(s, max_s, " ");
	_concat_ptr_to_function(func, s, max_s, t, MAX_T, flags);
	_concat(s, max_s, t);
	return;  /* Done, include the arguments. */
      } else
	_concat(s, max_s, " (*)(");
    }

    /* concat type arguments */
    _concat_type_list(func->args, s, max_s, flags);

    /* concat ")" */
    _concat(s, max_s, ")");
  }
  else /* ... print in the normal way */
  {
      struct	MLD_unmodified_type* local_ut;
      int       number_of_pointers = 1;
      int       i;
      /*
       * find out if this eventually points to an MLD_FUNCTION_TYPE. If so,
       * then do special processing right here so that all the *'s end up
       * between the ()'s.
       */
      local_ut = p->type_modified->ut;
      if (local_ut->shared_id) {
	_concat(s, max_s, local_ut->shared_id);
	DEMANGLE_TRACEOUT(_concat_pointer);
	return;
      }

      /* loop thru the pointers to pointers */
      while ((local_ut != 0) && (local_ut->type_of_type == MLD_POINTER_TYPE)) {
	local_ut = local_ut->the_type.pointer->type_modified->ut;
	number_of_pointers++;
      }

      if ((local_ut != 0) && (local_ut->type_of_type == MLD_FUNCTION_TYPE)) {
          _concat_type_structure(local_ut->the_type.function->return_type,
                          s, max_s, flags); 
          _concat(s, max_s, "(");

	  /* put in the *'s and pointer type modifiers */
          _concat_ptr_to_func_ptrs(p, s, max_s, flags);

          _concat (s, max_s, ")(");

          /* concat type arguments */
          _concat_type_list( local_ut->the_type.function->args, s,max_s,flags);

          _concat(s, max_s, ")");
      } else {

          /*
           * Need to do special processing for arrays of pointers to ensure
           * that the * comes after the type and before the dimensions.
           */
          if ((local_ut != 0) && (local_ut->type_of_type == MLD_ARRAY_TYPE)) {
              char temp[STRSIZ];
              struct MLD_unmodified_type* array_local_ut;

              array_local_ut = local_ut;

              /* loop thru the array indicies to find the pointer type */
              while ((array_local_ut != 0) &&
                  (array_local_ut->type_of_type == MLD_ARRAY_TYPE)  &&
                  (array_local_ut->the_type.array->type_modified->
                                ut->type_of_type == MLD_ARRAY_TYPE)) {
                  array_local_ut =
			array_local_ut->the_type.array->type_modified->ut;
              }

              /* concat the type, then the '*', then the [index][index]... */
	      _concat_type_structure(
                array_local_ut->the_type.array->type_modified, s, max_s, flags);
	      for (i=0; i<number_of_pointers; i++)
		_concat(s, max_s, "*");

              /* loop thru array indicies creating [num][num]etc. */
              while ((local_ut != 0) &&
                  (local_ut->type_of_type == MLD_ARRAY_TYPE)) {
                  _concat(s, max_s, "[");
                  _concat(s, max_s,
                    _itostr(temp, STRSIZ, local_ut->the_type.array->dimension));
                  _concat(s, max_s, "]");
                  local_ut = local_ut->the_type.array->type_modified->ut;
              }
          } else {
              /* concat type unmodified */
              _concat_unmodified_type(p->type_modified->ut, s, max_s, flags);

              if (p->type_modified) {
                  /* concat an asterisk, for the pointer */
		  //for (i=0; i<number_of_pointers; i++)
		    _concat(s, max_s, "* ");
                  _concat_type_modifier_list(p->type_modified->tm_list,
                                                       s, max_s, flags);
                 /* get rid of extra trailing blank */
                  s[strlen(s)-1] = 0;
              } else
		  for (i=0; i<number_of_pointers; i++)
		    _concat(s, max_s, "*");
          }
      }
  }
  DEMANGLE_TRACEOUT(_concat_pointer);
}


void
_dump_reference(struct MLD_reference* r)
{
  if (r == 0L) return;

  /* dump type modified */
  printf("(REF:");
  _dump_type_structure(r->type_modified);
  printf(":REF)\n");
}

void
_delete_reference(struct MLD_reference* r)
{
  if (r == 0L) return;

  DEMANGLE_TRACEIN(_delete_reference);

  /* delete type modified */
  _delete_type_structure(r->type_modified);
  
  /* delete self */
  DEMANGLE_FREE(r);
  DEMANGLE_TRACEOUT(_delete_reference);
}

void
_concat_reference(struct MLD_reference* r,
                 char* s,
                 int max_s, 
                 int flags)
{
  int unmodified_done = 0;
  if (r == 0L) return;

  DEMANGLE_TRACEIN(_concat_reference);

  /* concat type unmodified */
  if (r->type_modified && r->type_modified->ut &&
      r->type_modified->ut->shared_id)
    _concat (s, max_s, r->type_modified->ut->shared_id);
  else
    _concat_unmodified_type(r->type_modified->ut, s, max_s, flags);

  if (r->type_modified) {
    /* concat an ampersand, for the reference */
    _concat(s, max_s, "& ");
    _concat_type_modifier_list(r->type_modified->tm_list, s, max_s, flags);
    /* get rid of extra trailing blank */
    s[strlen(s)-1] = 0;
  } else
    _concat(s, max_s, "&");
  DEMANGLE_TRACEOUT(_concat_reference);

}


void
_dump_array(struct MLD_array* a)
{
  if (a == 0L) return;

  printf("(ARY[%d]:",a->dimension);

  /* dump type modified */
  _dump_type_structure(a->type_modified);
  printf(":YRA)\n");
}

void
_delete_array(struct MLD_array* a)
{
  if (a == 0L) return;

  DEMANGLE_TRACEIN(_delete_array);

  /* delete type modified */
  _delete_type_structure(a->type_modified);
  if (a->nontype_dim)
    _delete_type_structure(a->nontype_dim);
  /* delete self */
  DEMANGLE_FREE(a);
  DEMANGLE_TRACEOUT(_delete_array);
}

void
_concat_array(struct MLD_array* a,
             char* s,
             int max_s, 
             int flags)
{
  char temp[STRSIZ];
  if (a == 0L) return;

  DEMANGLE_TRACEIN(_concat_array);

  /* concat type modified */
  _concat_type_structure(a->type_modified, s, max_s, flags);
  
  /* concat dimension inside of square brackets */
  _concat(s, max_s, "[");
  if (a->nontype_dim) {
    _concat_type_structure(a->nontype_dim, s, max_s, flags);
  } else
    _concat(s, max_s, _itostr(temp, STRSIZ, a->dimension));
  _concat(s, max_s, "]");
  DEMANGLE_TRACEOUT(_concat_array);
}


void
_dump_pointer_to_member(struct MLD_pointer_to_member* pm)
{
  if (pm == 0L) return;

  /* dump qualifications */
  if (pm->qualifications != 0L)
    _dump_qualifications(pm->qualifications);
  
  /* dump type modified */
  _dump_type_structure(pm->type_modified);
}


void
_delete_pointer_to_member(struct MLD_pointer_to_member* pm)
{
  if (pm == 0L) return;

  DEMANGLE_TRACEIN(_delete_pointer_to_member);

  /* delete qualifications */
  _delete_qualifications(pm->qualifications);
  
  /* delete type modified */
  _delete_type_structure(pm->type_modified);
  
  /* delete self */
  DEMANGLE_FREE(pm);
  DEMANGLE_TRACEOUT(_delete_pointer_to_member);
}


_concat_ptr_to_mem_ptrs(struct MLD_pointer_to_member* pm,
               char* s,
               int max_s,
               int flags)
{
  DEMANGLE_TRACEIN(_concat_ptr_to_mem_ptrs);

    if (pm->type_modified->ut->type_of_type != MLD_FUNCTION_TYPE)
	_concat_ptr_to_func_ptrs (pm->type_modified->ut->the_type.pointer,
		s, max_s, flags);

    if (pm->qualifications != 0L)
    {
      _concat_qualifications(pm->qualifications, s, max_s, flags);
    }

    /* concatenate "::*" */
    _concat(s, max_s, "::* ");
    _concat_type_modifier_list(pm->type_modified->tm_list, s, max_s, flags);
  DEMANGLE_TRACEOUT(_concat_ptr_to_mem_ptrs);
}


void
_concat_pointer_to_member(struct MLD_pointer_to_member* pm,
                         char* s,
                         int max_s, 
                         int flags)
{
  if (pm == 0L) return;

  DEMANGLE_TRACEIN(_concat_pointer_to_member);

  /* if this is a pointer to a function, print in a special way... */
  if (pm->type_modified->ut->type_of_type == MLD_FUNCTION_TYPE)
  {
    /* concatenate return type */
    _concat_type_structure(
	pm->type_modified->ut->the_type.function->return_type, s, max_s, flags);

    /* concatenate " (" */
    _concat(s, max_s, " (");

    /* concat qualifications ("member")*/
    if (pm->qualifications != 0L)
    {
      _concat_qualifications(pm->qualifications, s, max_s, flags);
    }

    /* concatenate "::*" */
    _concat(s, max_s, "::* ");

    _concat_type_modifier_list(pm->type_modified->tm_list, s, max_s, flags);
    /* get rid of extra trailing blank */
    s[strlen(s)-1] = 0;

    _concat(s, max_s, ")(");

    /* concat type arguments */
    _concat_type_list(pm->type_modified->ut->the_type.function->args,
							s, max_s, flags); 
    /* concatenate ")" */
    _concat(s, max_s, ")");
    _concat_type_modifier_list(pm->type_modified->ut->the_type.function->tm_list,
			       s, max_s, flags);
    /* get rid of extra trailing blank */
    if (s[strlen(s)-1] == ' ')
      s[strlen(s)-1] = 0;
  }
  else /* ... print in the normal way */
  {

     struct	MLD_unmodified_type* local_ut;
     /*
      * find out if this eventually points to an MLD_FUNCTION_TYPE. If so,
      * then do special processing right here so that all the *'s end up
      * between the ()'s.
      */
     local_ut = pm->type_modified->ut;
     if (local_ut->shared_id) {
       _concat(s, max_s, local_ut->shared_id);
       DEMANGLE_TRACEOUT(_concat_pointer_to_member);
       return;
     }

     /* loop thru the pointers to pointers */
     while ((local_ut != 0) && (local_ut->type_of_type == MLD_POINTER_TYPE)) {
        local_ut = local_ut->the_type.pointer->type_modified->ut;
     }
     if ((local_ut != 0) && (local_ut->type_of_type == MLD_FUNCTION_TYPE)) {
          _concat_type_structure(local_ut->the_type.function->return_type,
                          s, max_s, flags); 
          _concat(s, max_s, "(");

	  /* put in the *'s and pointer type modifiers */
          _concat_ptr_to_mem_ptrs(pm, s, max_s, flags);

          _concat (s, max_s, ")(");

          /* concat type arguments */
          _concat_type_list( local_ut->the_type.function->args, s,max_s,flags);

          _concat(s, max_s, ")");
      } else {

         _concat_unmodified_type(pm->type_modified->ut, s, max_s, flags);
         _concat(s, max_s, " ");
         /* concat qualifications ("member")*/
         if (pm->qualifications != 0L)
         {
           _concat_qualifications(pm->qualifications, s, max_s, flags);
         }
 
         /* concat asterisk ("pointer") */
         _concat(s, max_s, "::* ");
 
         _concat_type_modifier_list(pm->type_modified->tm_list, s,max_s,flags);
         /* get rid of extra trailing blank */
         s[strlen(s)-1] = 0;
      }
  }
  DEMANGLE_TRACEOUT(_concat_pointer_to_member);
}


void
_dump_function(struct MLD_function* f)
{
  /* dump arguments */
  printf("(FUN:(ARG:");
  _dump_type_list(f->args);
  printf(":GRA) (RTN:");
  /* dump return type */
  _dump_type_structure(f->return_type);
  printf(":NTR) (MOD:");
  _dump_type_modifier_list(f->tm_list);
  printf(":DOM) :NUF)");
}

void
_delete_function(struct MLD_function* f)
{
  if (f == 0L) return;

  DEMANGLE_TRACEIN(_delete_function);

  /* delete arguments */
  _delete_type_list(f->args);
  
  /* delete return type */
  _delete_type_structure(f->return_type);

  _delete_function(f->ptr_to_function_type);
  _delete_type_modifier_list(f->tm_list);

  /* delete self */
  DEMANGLE_FREE(f);
  DEMANGLE_TRACEOUT(_delete_function);
}

void
_concat_function(struct MLD_function* f,
                char* s,
                int max_s, 
                int flags)
{
  DEMANGLE_TRACEIN(_concat_function);

  if (f->extern_c)
    _concat(s, max_s, "extern \"C\" ");

  /* concat return type */
  _concat_type_structure(f->args, s, max_s, flags);  
  _concat(s, max_s, " ");

  /* concat arguments */
  if (f->args != 0L) /* if there are some arguments */
  {
    _concat(s, max_s, "(");
    _concat_type_list(f->args, s, max_s, flags);
    _concat(s, max_s, ")");
  }
  DEMANGLE_TRACEOUT(_concat_function);
}


int /* recognition status */
_get_class(struct MLD_qualification* q,	/* qualification */
          struct MLD_complex** c) 		/* class */
{
  struct MLD_qualification* current_q, *next_q;


  /* check to see if q is empty */
  if (q == 0L)
  {
    return MLD_NOTHING_RECOGNIZED;
  }
  
  DEMANGLE_TRACEIN(_get_class);

  current_q = q; next_q = current_q->next;
  while (next_q != 0L)
  {
    current_q = next_q; next_q = current_q->next;
  }
  
  /* At this point current_q points to last the last qualification, the class */
  *c = current_q->q;

  DEMANGLE_TRACEOUT(_get_class);
  
  return MLD_SOMETHING_RECOGNIZED;
}

int /* number of double underscores found */
_num_dbl_uscores(const char* s,	/* IN:  string to be parsed */
          	 int start)	/* IN:  index of first unparsed character */
{
  int current_num_du = 0;
  
  char ch; int p;
    
  p = start;

  /* handle case where string starts with "__" */
  if ((s[p]=='_') && (s[p+1]=='_'))
  {
    p = p + 2; /* skip double underscore */
  }

  while(s[p] != 0) /* null terminates string */
  {
    if ((s[p]=='_') && (s[p+1]=='_')) /* double underscore found */
    {
      /* increment number of double underscores */
      current_num_du++;
  
      /* advance past the double underscores */
      while(s[p]=='_') p++;
    }
    else
    {
      p++;
    }
  }
  return current_num_du;
}

int /* recognition status */
demangle_unnamed(const char *s, /* IN: string to be parsed */
		 char *string)	/* OUT: of unnamed... */
{
  int i;
  char* unnamed_tokens[NUMBER_OF_UNNAMED_TOKENS+1] = UNNAMED_TOKENS;
  char* unnamed_strings[NUMBER_OF_UNNAMED_TOKENS+1] = UNNAMED_STRINGS;

  for (i = 0; i < NUMBER_OF_UNNAMED_TOKENS; i++)
    if (strncmp(s, unnamed_tokens[i], UNNAMED_TOKEN_LEN) == 0 &&
	(strncmp(s, UNNAMED_NAMESPACE_TOKEN, UNNAMED_TOKEN_LEN) == 0
	 || (s[UNNAMED_TOKEN_LEN] >= '1' && 
	     s[UNNAMED_TOKEN_LEN] <= '9'))) {
      /* Note, in __?nnn format, nnn could not have a leading zero.
       * Under model ansi, the unnamed namespace shall follow a module
       * name. i.e. __Nt_cxx_main where t.cxx(main) is the module name.
       */
      strcpy(string, unnamed_strings[i]);
      return MLD_SOMETHING_RECOGNIZED;
    }
  return MLD_NOTHING_RECOGNIZED;

}


#if !PRODUCTION
void* malloc_array[65536];

int demangle_trace_valid;
int demangle_trace_enable;

int demangle_memcheck;
int demangle_dump;
int my_index;
int my_pin_point_malloc;

#define DEMANGLE_DEV 1
#define DEMANGLE_CHECK_ADDR 0x140ad480
#define DEMANGLE_CHECK_MALLOC_POINT 8730

void demangle_check_free(void* foo) {
  int my_i;
  for (my_i=0; my_i<=my_index; my_i++) {
    if (malloc_array[my_i] == foo) {
      /* fill in the emptied element with the last element to keep the list
	 short. */
      malloc_array[my_i] = malloc_array[--my_index];
      malloc_array[my_index] = NULL;
#if 0
      if (DEMANGLE_DEV && foo == (void*) DEMANGLE_CHECK_ADDR) {
	printf ("%d  free: %0x\n", my_pin_point_malloc, foo);
      }
#endif
      break;
    } else if (my_i == my_index) {
      printf("ERROR demangle=========> Unexpected free at: %p\n",foo);
    }
  }
  return;
}

void demangle_dump_malloc(void* foo) {
#if 0
  if (DEMANGLE_DEV && foo == (void*) DEMANGLE_CHECK_ADDR) {
    my_pin_point_malloc++;
    printf ("%d  malloc: %0x\n", my_pin_point_malloc, foo);
    if (my_pin_point_malloc == DEMANGLE_CHECK_MALLOC_POINT)
      printf ("Found where the unfreed malloc is******************\n");
  }
#endif
    return;
}
#undef DEMANGLE_CHECK_ADDR
#undef DEMANGLE_DEV
#endif
