#pragma required_pointer_size long
#ifndef lint
static const char sccsid[] = "@(#)virtual_table_name.c	1.1     2/10/94";
#endif
/*
*****************************************************************************
**                                                                          *
**                         COPYRIGHT (c) 1991, 1996 BY                      *
**             DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASS.                *
**                          ALL RIGHTS RESERVED                             *
**                                                                          *
**  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND  COPIED  *
**  ONLY  IN  ACCORDANCE  WITH  THE  TERMS  OF  SUCH  LICENSE AND WITH THE  *
**  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR  ANY  OTHER  *
**  COPIES  THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY  *
**  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE  IS  HEREBY  *
**  TRANSFERRED.                                                            *
**                                                                          *
**  THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE  WITHOUT  NOTICE  *
**  AND  SHOULD  NOT  BE  CONSTRUED  AS  A COMMITMENT BY DIGITAL EQUIPMENT  *
**  CORPORATION.                                                            *
**                                                                          *
**  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE OR  RELIABILITY  OF  ITS  *
**  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
**                                                                          *
*****************************************************************************
**++
**  PRODUCT:
***
**	DEC C++ Demangler
**
**  FILE:  
**
**      virtual_table_name.c
**
**  ABSTRACT:
**
**	Grammar element that parses a mangled_name.
**
**  ENVIRONMENT:
**
**	User Mode
**
**  AUTHORS:
**
** 	Shell S. Simpson (DECwest Engineering)
**
**  MODIFICATION HISTORY:
**
**	 8-Jul-1996	Harold Seigel
**      Check lengths before doing strcpy's to prevent overwriting the stack.
**
**	29-Oct-1997	Shi-Jung Kao
**	Support __T_ for runtime type into.
**
**	16-Dec-1997	Shi-Jung Kao	cxxc_bugs: 5004
**	Parse_qualification.
**
**	05-Feb-1998	Shi-Jung Kao
**	Allow namespace at the end of the vtbl name.
**
**	08-Jan-1998	Shi-Jung Kao
**	Interpret __init_sti__ and __fini_sti__.
**
**	20-Jan-2000	Shi-Jung Kao
**	Accept the new model ansi vtbl mangled names.
**	e.g. __7__vtbl_8Derived3_8Derived3__8Derived4
**	Display the original vtbl name, instead of std::__vtbl_3std1C
**	Accept the __VTT__ for vtable too.
**--
**/


#ifdef CXX_NT
#include "defs.h"
#include "mn_nm.h"
#include "gram_elm.h"
#include "util.h"
#else
#include "demangle.h"
#include "grammar_elements.h"
#endif

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

/* language element particular definitions */
#define ASCII_ZERO 48
#define INTERNAL_INIT_TAIL 17

// Get the file name from __init_sti__ and __fini_sti__.
//
static void _get_file_name(const char* s, char* str_replaced_by_name) {
  char* file_suffix;
  if (file_suffix = strstr(s, "_cpp_")) {
    int i = file_suffix - s;
    strncpy(str_replaced_by_name, s, i);
    strncpy(&str_replaced_by_name[i], ".cpp", 4);
    str_replaced_by_name[i+4] = 0;
  } else if (file_suffix = strstr(s, "_cxx_")) {
    int i = file_suffix - s;
    strncpy(str_replaced_by_name, s, i);
    strncpy(&str_replaced_by_name[i], ".cxx", 4);
    str_replaced_by_name[i+4] = 0;
  } else if (file_suffix = strstr(s, "_C_")) {
    int i = file_suffix - s;
    strncpy(str_replaced_by_name, s, i);
    strncpy(&str_replaced_by_name[i], ".C", 2);
    str_replaced_by_name[i+2] = 0;
  } else if (strlen(s) > INTERNAL_INIT_TAIL) {
    int i = strlen(s) - INTERNAL_INIT_TAIL;
    strncpy(str_replaced_by_name, s, i);
    str_replaced_by_name[i] = 0;
  }
}

int /* recognition status */
_virtual_table_name
	(const char* s,		/* IN:  string to be parsed */
     int start,			/* IN:  index of first unparsed character */
     int* end,			/* OUT: index of last parsed character */
     char* message,		/* OUT: struct containing mangled funct name */
     char* str_before_name,	/* OUT: string before mangled funct name */
     char* str_replaced_by_name,/* OUT: string replaced by mangled funct name*/
     char* str_after_name,	/* OUT: string after mangled funct name */
     int size_of_string,	/* IN:	max size of the strings listed above */
     enum MLD_mangled_name_types *mld_type)	
				/* OUT:	max size of the strings listed above */
{
  /* variables common to all language elements */
  int ci;  /* current index */
  char ch; /* current character */
  
  /* variables particular to this language element */
  int status;
  int end_of_grammar_component;
  char qual1[STRSIZ];		/* qualification name */
  char qual2[STRSIZ];		/* qualification name */
  char class[STRSIZ];		/* class name */
  char temp[STRSIZ];		/* temporary string */
  int number_recognized = 0;	/* number of pascal strings recognized */
  int end_of_mn;		/* index of last character in s of mn */
  int end_of_mn_prenamespace = 0;/* index of last character in s of mn */
  struct MLD_qualification* q=0;/* qualifier */  

  /* initialization particular to language element */
  *end = start - 1;
  if ((strlen(s) - start) < size_of_string) { /* copy if enough room */
    strcpy(str_before_name, &s[start]); 
  }     
  if (size_of_string > 0) { /* copy if enough room */
    strcpy(str_replaced_by_name,"");
  }    
  if (size_of_string > 0) { /* copy if enough room */
    strcpy(str_after_name,"");
  }
  
  ch = s[ci = start]; /* initialize ci and ch */
 
  /*
  **
  ** Does s begin with "__VTT__", "__vtbl_" or "__btbl_" ?
  **
  */
  
  if ((strncmp(&s[start], "__VTT__", 7) == 0) ||
	    (strncmp(&s[start], "__vtbl_", 7) == 0) ||
	    (strncmp(&s[start], "__btbl_", 7) == 0)) {
    ch = s[ci = (ci + 7)]; /* initialize ci and ch */
    *mld_type = MLD_VIRTUAL_TABLE_NAME;

  } else if (strncmp(&s[start], "__T_", 4) == 0) {
    ch = s[ci = (ci + 4)]; /* initialize ci and ch */
    *mld_type = MLD_RUNTIMETYPE_INFO_NAME;

  } else if (strncmp(&s[start], "__init_sti__", 12) == 0) {
    char file[STRSIZ];		/* file name */
  
    ch = s[ci = (ci + 12)]; /* initialize ci and ch */
    *mld_type = MLD_INIT_INTERNAL_NAME;

    file[0] = 0;
    _get_file_name(&s[ci],file); 
    if (file[0] == 0)
      return MLD_NOTHING_RECOGNIZED;
    strcpy(message, file);
    strcpy(str_before_name, ""); 

    *end = strlen(s);
    return MLD_SOMETHING_RECOGNIZED;

  } else if (strncmp(&s[start], "__fini_sti__", 12) == 0) {
    char file[STRSIZ];		/* file name */

    ch = s[ci = (ci + 12)]; /* initialize ci and ch */
    *mld_type = MLD_FINI_INTERNAL_NAME;

    file[0] = 0;
    _get_file_name(&s[ci],file);
    if (file[0] == 0)
      return MLD_NOTHING_RECOGNIZED;
    strcpy(message, file);
    strcpy(str_before_name, ""); 

    *end = strlen(s);
    return MLD_SOMETHING_RECOGNIZED;

  } else {
    /* ...nothing recognized */
    return MLD_NOTHING_RECOGNIZED;
  } 

  /*
  **
  ** Get the qualifications cxxc_bugs: 5004
  **
  */
  qual1[0] = '\0';
  qual2[0] = '\0';
  class[0] = '\0';
  status = _parse_qualifications(s, ci, &end_of_grammar_component, &q);

  switch (status) {    
  case MLD_NOTHING_RECOGNIZED:
      /* no problem */
      q = 0L; /* make sure that qualification list is empty */
      break;
    
  case MLD_ERROR:
      /* if error occured, return ERROR */
      *end = start - 1; /* nothing recognized */
      //*mfn = 0L; /* no mangled name */
      
      if (size_of_string > 0) /* copy if enough room */
      {
        strcpy(str_before_name,"");
        strcpy(str_replaced_by_name,"");
      }
      
      if ((strlen(s) - start) < size_of_string) /* copy if enough room */
      {
        strcpy(str_after_name, &s[start]); /* everything is "after" */
      }
      
      return MLD_ERROR;
      break;

  case MLD_SOMETHING_RECOGNIZED:
      qual1[0] = '\0';
      _concat_qualifications(q, &qual1[0], STRSIZ, 0/*flags*/);
      number_recognized++;
      break;

  default:
      break;
  }
  ch = s[ci = (end_of_grammar_component+1)]; /* update ci and ch */
  /* deallocate memory allocated by function _parse_qualifications() */
  _delete_qualifications(q);
  q = 0L;

  /*
  **
  ** Keep getting pascal strings until no more are recognized
  ** (The last one is the name of the class)
  **
  */
  
  if (*mld_type == MLD_VIRTUAL_TABLE_NAME && 
      MLD_mangled_version == MLD_VERSION7) {
    /* Skip the underscore. At the most 2 of it. */
    if (ch == '_')
      ch = s[++ci];
    if (ch == '_') {
      /* Under version7 vtbl, this must be the last one. */
      ch = s[++ci];
      qual1[0] = '\0';
    }
  }
  
  while((status = 
		_pascal_string(s, ci, &end_of_grammar_component, temp, STRSIZ))
	== MLD_SOMETHING_RECOGNIZED) {
    number_recognized++;
    ch = s[ci = (end_of_grammar_component+1)]; /* update ci and ch */
    if (*mld_type == MLD_VIRTUAL_TABLE_NAME &&
	MLD_mangled_version == MLD_VERSION7) {
      /* Skip the underscore. (At the most 2 of it.) */
      if (ch == '_')
	ch = s[++ci];
      if (ch == '_') {
	/* Under version7 vtbl, this must be the last one. */
	ch = s[++ci];
	qual1[0] = '\0';
      }
    }
    if (strlen(temp) < STRSIZ) {
      /* Under version7 vtbl, ignore all classes but the last one.
       * It is assumed, that vtbl with namespace will never goes here.
       * e.g. __7__vtbl_Q2_3std9type_info
       */
      if (*mld_type == MLD_VIRTUAL_TABLE_NAME &&
	  MLD_mangled_version == MLD_VERSION7)
	qual1[0] = '\0';

      strcpy(class, temp);
    }
  }
  if (status == MLD_ERROR) {
    return MLD_ERROR;
  }
  if (number_recognized == 0) {
    return MLD_NOTHING_RECOGNIZED;
  }

  if (strlen(s) >= ci+2 && (s[ci] == '_') && s[ci+1] == '_') {
    /* Check there might me a namespace at the end. */
    ch = s[ci = (ci+2)]; /* update ci and ch */
    status = _parse_qualifications(s, ci, &end_of_grammar_component, &q);
    switch (status) {    
    case MLD_NOTHING_RECOGNIZED:
      /* no problem */
      q = 0L; /* make sure that qualification list is empty */
      break;
    
    case MLD_ERROR:
      /* if error occured, return ERROR */
      *end = start - 1; /* nothing recognized */
      //*mfn = 0L; /* no mangled name */
      
      if (size_of_string > 0) /* copy if enough room */
      {
        strcpy(str_before_name,"");
        strcpy(str_replaced_by_name,"");
      }
      
      if ((strlen(s) - start) < size_of_string) /* copy if enough room */
      {
        strcpy(str_after_name, &s[start]); /* everything is "after" */
      }
     
      return MLD_ERROR;
      break;

    case MLD_SOMETHING_RECOGNIZED:
      qual2[0] = '\0';
      _concat_qualifications(q, &qual2[0], STRSIZ, 0/*flags*/);
      number_recognized++;
      end_of_mn_prenamespace = ci - 2;
      break;

    defalut:
      break;
    }
    ch = s[ci = (end_of_grammar_component+1)]; /* update ci and ch */
    /* deallocate memory allocated by function _parse_qualifications() */
    _delete_qualifications(q);
    q = 0L;
  }

  /*
  **
  ** Check for white space
  **
  */
  
  end_of_mn = end_of_grammar_component;
  status = _white_space(s, end_of_grammar_component + 1, 
					&end_of_grammar_component);
  switch (status) {    
    case MLD_NOTHING_RECOGNIZED:
      return MLD_NOTHING_RECOGNIZED;
    break;
    
    case MLD_ERROR:
      /* if error occured, return ERROR */            
      return MLD_ERROR;
    break;
  }
 
  /*
  **
  ** All the trials have been passed.  Now is the type to create the
  ** message string.
  **
  */
  
  if (qual1[0] != '\0' && strlen(qual1) < size_of_string) {
    if (class[0] != '\0' && strlen(class) < size_of_string) {
      _concat(qual1, size_of_string, "::");
      _concat(qual1, size_of_string, class);
    }
    strcpy(message, qual1); /* copy the qual1::class to the message string */
  } else if (strlen(class) < size_of_string) {
    strcpy(message, class); /* copy the class to the message string */
  }

  if (0 < size_of_string) { /* copy if enough room */
    strcpy(str_before_name, ""); 
  }
      
  if ((end_of_mn - start + 1) < size_of_string) { /* copy if enough room */
#if 0
    /* Code commented out because we want to
     * present the original vtbl name.
     * i.e. __vtbl_3std1C__3std instead of std::__vtbl_3std1C
     */

    if (qual2[0] != '\0') {
      /* This is the __vtbl_ under a namespace of qual2 */
      int string_len_upto_namespace;
      _concat(qual2, size_of_string, "::");
      strcpy(str_replaced_by_name, qual2);
      if (end_of_mn_prenamespace == 0)
	end_of_mn_prenamespace = end_of_mn;
      string_len_upto_namespace = 
	strlen(qual2) + end_of_mn_prenamespace - start;
      strncpy(&str_replaced_by_name[strlen(qual2)], 
	      &s[start], string_len_upto_namespace);
      /* null terminate */
      str_replaced_by_name[string_len_upto_namespace] = 0;
    } else {
#endif
      strncpy(str_replaced_by_name, &s[start], end_of_mn - start + 1 );
      str_replaced_by_name[(end_of_mn - start + 1)] = 0; /* null terminate */
#if 0
    }
#endif
  }
      
  if (strlen(&s[end_of_mn+1]) < size_of_string) { /* copy if enough room */
    strcpy(str_after_name, &s[end_of_mn+1]);
  }
      
  *end = end_of_mn; /* end of mangled name */
  return MLD_SOMETHING_RECOGNIZED;
}
